#!/usr/bin/env python3

from typing import List

import numpy as np

from .interp import linear_interp


def interp2d(
    Q: List[np.ndarray],
    dy: np.ndarray,
    dx: np.ndarray,
    mode: str = "bilinear",
) -> np.ndarray:
    r"""Naive Interpolation
    (y,x): target pixel
    mode: interpolation mode
    """
    q00, q10, q01, q11 = Q
    if mode == "bilinear":
        f0 = linear_interp(q00, q01, dx, 1)
        f1 = linear_interp(q10, q11, dx, 1)
        return linear_interp(f0, f1, dy, 1)
    else:
        raise NotImplementedError


def grid_sample(
    img: np.ndarray,
    grid: np.ndarray,
    mode: str = "bilinear",
) -> np.ndarray:
    r"""Numpy Grid Sample"""
    channels, h_in, w_in = img.shape
    _, h_out, w_out = grid.shape

    # Image conversion values (Lookup Image dtype)
    if img.dtype == np.uint8:
        # uint8 is faster
        _min = 0
        _max = 255
        _dtype = np.uint8
    elif img.dtype == np.float64:
        _min = 0.0
        _max = 1.0
        _dtype = np.float64
    else:
        raise ValueError("{} is not supported".format(img.dtype))

    # Initialize output image
    out = np.zeros((channels, h_out, w_out), dtype=_dtype)

    if mode == "bilinear":
        # NOTE: uint8 convertion causes truncation, so use uint64
        min_grid = np.floor(grid).astype(np.uint64)
        max_grid = min_grid + 1
        d_grid = grid - min_grid

        y_d = d_grid[0, :, :]
        x_d = d_grid[1, :, :]
        y_d = y_d.flatten()
        x_d = x_d.flatten()

        max_grid[0, :, :] = np.where(
            max_grid[0, :, :] >= h_in,
            max_grid[0, :, :] - h_in,
            max_grid[0, :, :],
        )
        max_grid[1, :, :] = np.where(
            max_grid[1, :, :] >= w_in,
            max_grid[1, :, :] - w_in,
            max_grid[1, :, :],
        )

        y_mins = min_grid[0, :, :]
        x_mins = min_grid[1, :, :]
        y_mins = y_mins.flatten()
        x_mins = x_mins.flatten()

        y_maxs = max_grid[0, :, :]
        x_maxs = max_grid[1, :, :]
        y_maxs = y_maxs.flatten()
        x_maxs = x_maxs.flatten()

        Q00 = img[:, y_mins, x_mins]
        Q10 = img[:, y_maxs, x_mins]
        Q01 = img[:, y_mins, x_maxs]
        Q11 = img[:, y_maxs, x_maxs]
        out = interp2d([Q00, Q10, Q01, Q11], y_d, x_d, mode=mode)

    elif mode == "nearest":
        round_grid = np.rint(grid).astype(np.uint64)
        y = round_grid[0, :, :]
        x = round_grid[1, :, :]
        y = y.flatten()
        x = x.flatten()

        out = img[:, y, x]

    else:
        raise ValueError("{} is not available".format(mode))

    out = np.where(out >= _max, _max, out)
    out = np.where(out < _min, _min, out)
    out = out.reshape(channels, h_out, w_out)
    return out.astype(_dtype)



#!/usr/bin/env python3

import math
from typing import Dict, List, Tuple, Union

import numpy as np

import torch

from equilib.grid_sample import torch_func

from ..base import BaseEqui2Equi
from .utils import create_rotation_matrix, get_device, sizeof

__all__ = ["Equi2Equi"]


class Equi2Equi(BaseEqui2Equi):
    def __init__(self, **kwargs):
        r"""Equi2Equi Numpy"""
        super().__init__(**kwargs)

    def create_coordinate(self, h_out: int, w_out: int) -> np.ndarray:
        r"""Create mesh coordinate grid with height and width

        return:
            coordinate: numpy.ndarray
        """
        xs = torch.linspace(0, w_out - 1, w_out)
        theta = xs * 2 * math.pi / w_out - math.pi
        ys = torch.linspace(0, h_out - 1, h_out)
        phi = ys * math.pi / h_out - math.pi / 2
        # NOTE: https://github.com/pytorch/pytorch/issues/15301
        # Torch meshgrid behaves differently than numpy
        phi, theta = torch.meshgrid([phi, theta])
        coord = torch.stack((theta, phi), axis=-1)
        return coord

    def rotation_matrix(
        self,
        roll: float,
        pitch: float,
        yaw: float,
    ) -> np.ndarray:
        r"""Create Rotation Matrix

        params:
            roll: x-axis rotation float
            pitch: y-axis rotation float
            yaw: z-axis rotation float

        return:
            rotation matrix: numpy.ndarray

        Global coordinates -> x-axis points forward, z-axis points upward
        """
        R = create_rotation_matrix(x=roll, y=pitch, z=yaw)
        return R

    @staticmethod
    def _get_img_size(img: torch.Tensor) -> Tuple[int]:
        r"""Return height and width"""
        # batch, channel, height, width
        return img.shape[-2:]

    def run(
        self,
        src: torch.Tensor,
        rot: Union[Dict[str, float], List[Dict[str, float]]],
        sampling_method: str = "torch",
        mode: str = "bilinear",
        debug: bool = False,
    ) -> torch.Tensor:
        r"""Run Equi2Pers

        params:
            src: equirectangular image torch.Tensor[(B), C, H, W]
            rot: Dict[str, float] or List[Dict[str, float]]
            sampling_method: str (default="torch")
            mode: str (default="bilinear")

        returns:
            pers: perspective image torch.Tensor[C, H, W]

        NOTE: input can be batched [B, C, H, W] or single [C, H, W]
        NOTE: when using batches, the output types match
        """
        assert type(src) == torch.Tensor, (
            "ERR: input equi expected to be `torch.Tensor` "
            "but got {}".format(type(src))
        )
        _original_shape_len = len(src.shape)
        assert _original_shape_len >= 3, "ERR: got {} for input equi".format(
            _original_shape_len
        )
        if _original_shape_len == 3:
            src = src.unsqueeze(dim=0)
            rot = [rot]

        h_equi, w_equi = self._get_img_size(src)
        if self.h_out is None and self.w_out is None:
            self.h_out = h_equi
            self.w_out = w_equi

        if debug:
            print("size of src: ", sizeof(src) / 10e6, "mb")

        # get device
        device = get_device(src)

        # define variables
        B = []
        for r in rot:
            a = self.create_coordinate(self.h_out, self.w_out)
            norm_A = 1
            x = norm_A * torch.cos(a[:, :, 1]) * torch.cos(a[:, :, 0])
            y = norm_A * torch.cos(a[:, :, 1]) * torch.sin(a[:, :, 0])
            z = norm_A * torch.sin(a[:, :, 1])
            A = torch.stack((x, y, z), dim=-1)
            R = self.rotation_matrix(**r)
            _B = R @ A.unsqueeze(3)
            _B = _B.squeeze(3)
            B.append(_B)
        B = torch.stack(B, dim=0).to(device)

        # calculate rotations per perspective coordinates
        norms = torch.norm(B, dim=-1)
        theta = torch.atan2(B[:, :, :, 1], B[:, :, :, 0])
        phi = torch.asin(B[:, :, :, 2] / norms)

        # center the image and convert to pixel locatio
        ui = (theta - math.pi) * w_equi / (2 * math.pi)
        uj = (phi - math.pi / 2) * h_equi / math.pi
        # out-of-bounds calculations
        ui = torch.where(ui < 0, ui + w_equi, ui)
        ui = torch.where(ui >= w_equi, ui - w_equi, ui)
        uj = torch.where(uj < 0, uj + h_equi, uj)
        uj = torch.where(uj >= h_equi, uj - h_equi, uj)
        grid = torch.stack((uj, ui), axis=-3)  # 3rd to last

        # grid sample
        grid_sample = getattr(torch_func, sampling_method, "torch")
        samples = grid_sample(src, grid, mode=mode)

        if _original_shape_len == 3:
            samples = samples.squeeze(axis=0)

        return samples



#!/usr/bin/env python3

import argparse
import os.path as osp
import time
from typing import Union

import cv2

import matplotlib
import matplotlib.pyplot as plt

import numpy as np

from PIL import Image

from equilib.equi2pers import NumpyEqui2Pers

matplotlib.use("Agg")


def preprocess(
    img: Union[np.ndarray, Image.Image],
    is_cv2: bool = False,
) -> np.ndarray:
    r"""Preprocesses image"""
    if isinstance(img, np.ndarray) and is_cv2:
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    if isinstance(img, Image.Image):
        # Sometimes images are RGBA
        img = img.convert("RGB")
        img = np.asarray(img)
    assert len(img.shape) == 3, "input must be dim=3"
    assert img.shape[-1] == 3, "input must be HWC"
    img = np.transpose(img, (2, 0, 1))
    return img


def test_video(
    path: str,
    h_pers: int = 480,
    w_pers: int = 640,
    fov_x: float = 90.0,
) -> None:
    r"""Test video"""
    # Rotation:
    pi = np.pi
    inc = pi / 180
    roll = 0  # -pi/2 < a < pi/2
    pitch = 0  # -pi < b < pi
    yaw = 0

    # Initialize equi2pers
    equi2pers = NumpyEqui2Pers(w_pers=w_pers, h_pers=h_pers, fov_x=fov_x)

    times = []
    cap = cv2.VideoCapture(path)

    while cap.isOpened():
        ret, frame = cap.read()

        rot = {
            "roll": roll,
            "pitch": pitch,
            "yaw": yaw,
        }

        if not ret:
            break

        s = time.time()
        equi_img = preprocess(frame, is_cv2=True)
        pers_img = equi2pers(
            equi=equi_img,
            rot=rot,
            sampling_method="faster",
            mode="bilinear",
        )
        pers_img = np.transpose(pers_img, (1, 2, 0))
        pers_img = cv2.cvtColor(pers_img, cv2.COLOR_RGB2BGR)
        e = time.time()
        times.append(e - s)

        # cv2.imshow("video", pers)

        # change direction `wasd` or exit with `q`
        k = cv2.waitKey(1)
        if k == ord("q"):
            break
        if k == ord("w"):
            roll -= inc
        if k == ord("s"):
            roll += inc
        if k == ord("a"):
            pitch += inc
        if k == ord("d"):
            pitch -= inc

    cap.release()
    cv2.destroyAllWindows()

    print(sum(times) / len(times))
    x_axis = [i for i in range(len(times))]
    plt.plot(x_axis, times)
    save_path = osp.join("./results", "times_equi2pers_numpy_video.png")
    plt.savefig(save_path)


def test_image(
    path: str,
    h_pers: int = 480,
    w_pers: int = 640,
    fov_x: float = 90.0,
) -> None:
    r"""Test single image"""
    # Rotation:
    rot = {
        "roll": 0,  #
        "pitch": np.pi / 4,  # vertical
        "yaw": np.pi / 4,  # horizontal
    }

    # Initialize equi2pers
    equi2pers = NumpyEqui2Pers(w_pers=w_pers, h_pers=h_pers, fov_x=fov_x)

    # Open Image
    equi_img = Image.open(path)
    equi_img = preprocess(equi_img)

    pers_img = equi2pers(
        equi_img,
        rot=rot,
        sampling_method="faster",
        mode="bilinear",
    )

    pers_img = np.transpose(pers_img, (1, 2, 0))
    pers_img = Image.fromarray(pers_img)

    pers_path = osp.join("./results", "output_equi2pers_numpy_image.jpg")
    pers_img.save(pers_path)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--video", action="store_true")
    parser.add_argument("--data", nargs="?", default=None, type=str)
    args = parser.parse_args()

    # Variables:
    h_pers = 480
    w_pers = 640
    fov_x = 90

    data_path = args.data
    if args.video:
        if data_path is None:
            data_path = "./data/R0010028_er_30.MP4"
        assert osp.exists(data_path)
        test_video(data_path, h_pers, w_pers, fov_x)
    else:
        if data_path is None:
            data_path = "./data/equi.jpg"
        assert osp.exists(data_path)
        test_image(data_path, h_pers, w_pers, fov_x)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

from .base import BaseEqui2Equi
from .equi2equi_numpy.equi2equi import Equi2Equi as NumpyEqui2Equi
from .equi2equi_torch.equi2equi import Equi2Equi as TorchEqui2Equi

__all__ = [
    "BaseEqui2Equi",
    "NumpyEqui2Equi",
    "TorchEqui2Equi",
]



#!/usr/bin/env python3

from typing import Dict, List, Tuple, Union

import numpy as np

from equilib.grid_sample import numpy_func

from ..base import BaseCube2Equi


def cube_list2h(cube_list: List[np.ndarray]):
    assert len(cube_list) == 6
    assert sum(face.shape == cube_list[0].shape for face in cube_list) == 6
    return np.concatenate(cube_list, axis=-1)


def cube_dict2h(
    cube_dict: Dict[str, np.ndarray], face_k=["F", "R", "B", "L", "U", "D"]
):
    assert len(face_k) == 6
    return cube_list2h([cube_dict[k] for k in face_k])


def cube_dice2h(cube_dice: np.ndarray):
    r"""dice to horizion

    params:
    cube_dice: (C, H, W)
    """
    # Order: F R B L U D
    sxy = [(1, 1), (2, 1), (3, 1), (0, 1), (1, 0), (1, 2)]
    w = cube_dice.shape[-2] // 3
    assert cube_dice.shape[-2] == w * 3 and cube_dice.shape[-1] == w * 4
    cube_h = np.zeros((cube_dice.shape[-3], w, w * 6), dtype=cube_dice.dtype)
    for i, (sx, sy) in enumerate(sxy):
        face = cube_dice[:, sy * w : (sy + 1) * w, sx * w : (sx + 1) * w]
        cube_h[:, :, i * w : (i + 1) * w] = face
    return cube_h


class Cube2Equi(BaseCube2Equi):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _to_horizon(self, cubemap: np.ndarray, cube_format: str) -> np.ndarray:
        if cube_format == "horizon":
            pass
        elif cube_format == "list":
            cubemap = cube_list2h(cubemap)
        elif cube_format == "dict":
            cubemap = cube_dict2h(cubemap)
        elif cube_format == "dice":
            cubemap = cube_dice2h(cubemap)
        else:
            raise NotImplementedError("unknown cube_format")

        assert len(cubemap.shape) == 3
        assert cubemap.shape[-2] * 6 == cubemap.shape[-1]

        return cubemap

    def _equirect_facetype(self, h: int, w: int):
        r"""0F 1R 2B 3L 4U 5D"""

        tp = np.roll(
            np.arange(4).repeat(w // 4)[None, :].repeat(h, 0), 3 * w // 8, 1
        )

        # Prepare ceil mask
        mask = np.zeros((h, w // 4), np.bool)
        idx = np.linspace(-np.pi, np.pi, w // 4) / 4
        idx = h // 2 - np.around(np.arctan(np.cos(idx)) * h / np.pi).astype(int)
        for i, j in enumerate(idx):
            mask[:j, i] = 1
        mask = np.roll(np.concatenate([mask] * 4, 1), 3 * w // 8, 1)

        tp[mask] = 4
        tp[np.flip(mask, 0)] = 5

        return tp.astype(np.int32)

    def create_equi_grid(self, h_out: int, w_out: int) -> Tuple[np.ndarray]:
        _dtype = np.float32
        theta = np.linspace(-np.pi, np.pi, num=w_out, dtype=_dtype)
        phi = np.linspace(np.pi, -np.pi, num=h_out, dtype=_dtype) / 2
        theta, phi = np.meshgrid(theta, phi)
        return theta, phi

    def _run_single(
        self,
        cubemap: np.ndarray,  # `horizon`
        sampling_method: str = "faster",
        mode: str = "bilinear",
    ) -> np.ndarray:
        r"""Run a single batch of transformation"""
        w_face = cubemap.shape[-2]

        theta, phi = self.create_equi_grid(self.h_out, self.w_out)

        # Get face id to each pixel: 0F 1R 2B 3L 4U 5D
        tp = self._equirect_facetype(self.h_out, self.w_out)

        # xy coordinate map
        coor_x = np.zeros((self.h_out, self.w_out))
        coor_y = np.zeros((self.h_out, self.w_out))

        for i in range(6):
            mask = tp == i

            if i < 4:
                coor_x[mask] = 0.5 * np.tan(theta[mask] - np.pi * i / 2)
                coor_y[mask] = (
                    -0.5
                    * np.tan(phi[mask])
                    / np.cos(theta[mask] - np.pi * i / 2)
                )
            elif i == 4:
                c = 0.5 * np.tan(np.pi / 2 - phi[mask])
                coor_x[mask] = c * np.sin(theta[mask])
                coor_y[mask] = c * np.cos(theta[mask])
            elif i == 5:
                c = 0.5 * np.tan(np.pi / 2 - np.abs(phi[mask]))
                coor_x[mask] = c * np.sin(theta[mask])
                coor_y[mask] = -c * np.cos(theta[mask])

        # Final renormalize
        coor_x = np.clip(np.clip(coor_x + 0.5, 0, 1) * w_face, 0, w_face - 1)
        coor_y = np.clip(np.clip(coor_y + 0.5, 0, 1) * w_face, 0, w_face - 1)

        # change x axis of the x coordinate map
        for i in range(6):
            mask = tp == i
            coor_x[mask] = coor_x[mask] + w_face * i

        grid = np.stack((coor_y, coor_x), axis=0)
        grid_sample = getattr(numpy_func, sampling_method)
        equi = grid_sample(cubemap, grid, mode=mode)

        return equi

    def run(
        self,
        cubemap: Union[np.ndarray, Dict[str, np.ndarray], List[np.ndarray]],
        cube_format: str = "dice",
        sampling_method: str = "faster",
        mode: str = "bilinear",
    ) -> np.ndarray:
        r"""Run cube to equirectangular image transformation

        params:
            cubemap: np.ndarray
            cube_format: ('dice', 'horizon', 'list', 'dict')
            sampling_method: str
            mode: str
        """

        # Convert all cubemap format to `horizon` and batched
        is_single = False
        if cube_format in ["dice", "horizon"]:
            if isinstance(cubemap, np.ndarray):
                # could be single or batched
                assert (
                    len(cubemap.shape) >= 3
                ), "input shape {} is not valid".format(cubemap.shape)
                if len(cubemap.shape) == 4:
                    # batch
                    cubemap = [
                        self._to_horizon(c, cube_format) for c in cubemap
                    ]
                else:
                    # single
                    is_single = True
                    cubemap = [self._to_horizon(cubemap, cube_format)]
            elif isinstance(cubemap, list):
                # could only be batched
                cubemap = [self._to_horizon(c, cube_format) for c in cubemap]
            else:
                raise ValueError
        elif cube_format == "dict":
            if isinstance(cubemap, dict):
                # single
                is_single = True
                cubemap = [self._to_horizon(cubemap, cube_format)]
            elif isinstance(cubemap, list):
                # batch
                cubemap = [self._to_horizon(c, cube_format) for c in cubemap]
            else:
                raise ValueError
        elif cube_format == "list":
            if isinstance(cubemap[0], np.ndarray):
                # single
                is_single = True
                cubemap = [self._to_horizon(cubemap, cube_format)]
            elif isinstance(cubemap[0], list):
                # batch
                cubemap = [self._to_horizon(c, cube_format) for c in cubemap]
            else:
                raise ValueError
        else:
            raise ValueError

        equis = []
        for cm in cubemap:
            equi = self._run_single(
                cm,
                sampling_method=sampling_method,
                mode=mode,
            )
            equis.append(equi)

        equis = np.stack(equis, axis=0)

        if is_single:
            equis = equis.squeeze(0)

        return equis



#!/usr/bin/env python3

from setuptools import Extension, setup

from torch.utils import cpp_extension


setup(
    name="grid_sample_cpp",
    ext_modules=[
        cpp_extension.CppExtension("grid_sample", ["grid_sample.cpp"])
    ],
    cmdclass={"build_ext": cpp_extension.BuildExtension},
)

Extension(
    name="grid_sample_cpp",
    sources=["grid_sample.cpp"],
    include_dirs=cpp_extension.include_paths(),
    language="c++",
)



#!/usr/bin/env python3

from typing import Dict, Tuple

import numpy as np

import torch

pi = torch.Tensor([3.14159265358979323846])


def sizeof(tensor: torch.Tensor) -> float:
    r"""Get the size of a tensor"""
    assert torch.is_tensor(tensor), "ERR: is not tensor"
    return tensor.element_size() * tensor.nelement()


def get_device(a: torch.Tensor) -> torch.device:
    r"""Get device of a Tensor"""
    return torch.device(a.get_device() if a.get_device() >= 0 else "cpu")


def deg2rad(tensor: torch.Tensor) -> torch.Tensor:
    r"""Function that converts angles from degrees to radians."""
    if not torch.is_tensor(tensor):
        return tensor * float(pi) / 180.0
    return tensor * pi.to(tensor.device).type(tensor.dtype) / 180.0


def create_M(
    height: int,
    width: int,
    fov_x: float,
    rot: Dict[str, float],
) -> torch.Tensor:
    r"""Create M"""
    m = create_coord(height, width)
    K = create_K(height, width, fov_x)
    R = create_rot_mat(**rot)
    M = R.inverse() @ K.inverse() @ m.unsqueeze(3)
    M = M.squeeze(3)
    return M


def create_coord(
    height: int,
    width: int,
) -> torch.Tensor:
    r"""Create mesh coordinate grid"""
    _xs = torch.linspace(0, width - 1, width)
    _ys = torch.linspace(0, height - 1, height)
    # NOTE: https://github.com/pytorch/pytorch/issues/15301
    # Torch meshgrid behaves differently than numpy
    ys, xs = torch.meshgrid([_ys, _xs])
    zs = torch.ones_like(xs)
    coord = torch.stack((xs, ys, zs), dim=2)
    return coord


def create_K(
    height: int,
    width: int,
    fov_x: float,
    skew: float = 0.0,
) -> torch.Tensor:
    r"""Create Intrinsic Matrix

    params:
        height: int
        width: int
        fov_x: float
        skew: float

    return:
        K: 3x3 matrix torch.Tensor

    NOTE:
        ref: http://ksimek.github.io/2013/08/13/intrinsic/
    """
    fov_x = torch.tensor(fov_x)
    f = width / (2 * torch.tan(deg2rad(fov_x) / 2))
    K = torch.tensor(
        [[f, skew, width / 2], [0.0, f, height / 2], [0.0, 0.0, 1.0]]
    )
    return K


def create_rotation_matrix(
    x: float,
    y: float,
    z: float,
) -> torch.Tensor:
    r"""Create Rotation Matrix

    params:
        x: x-axis rotation float
        y: y-axis rotation float
        z: z-axis rotation float

    return:
        rotation matrix: torch.Tensor
    """
    # calculate rotation about the x-axis
    R_x = torch.tensor(
        [
            [1.0, 0.0, 0.0],
            [0.0, np.cos(x), -np.sin(x)],
            [0.0, np.sin(x), np.cos(x)],
        ],
        dtype=torch.float,
    )
    # calculate rotation about the y-axis
    R_y = torch.tensor(
        [
            [np.cos(y), 0.0, np.sin(y)],
            [0.0, 1.0, 0.0],
            [-np.sin(y), 0.0, np.cos(y)],
        ],
        dtype=torch.float,
    )
    # calculate rotation about the z-axis
    R_z = torch.tensor(
        [
            [np.cos(z), -np.sin(z), 0.0],
            [np.sin(z), np.cos(z), 0.0],
            [0.0, 0.0, 1.0],
        ],
        dtype=torch.float,
    )

    return R_z @ R_y @ R_x


def create_rot_mat(
    roll: float,
    pitch: float,
    yaw: float,
) -> torch.Tensor:
    r"""Create Rotation Matrix

    params:
        roll: x-axis rotation float
        pitch: y-axis rotation float
        yaw: z-axis rotation float

    return:
        rotation matrix: torch.Tensor

    Camera coordinates -> z-axis points forward, y-axis points upward
    Global coordinates -> x-axis points forward, z-axis poitns upward

    NOTE: https://www.sciencedirect.com/topics/engineering/intrinsic-parameter
    """
    # default rotation that changes global to camera coordinates
    x = np.pi
    y = np.pi
    z = np.pi
    R = create_rotation_matrix(x=x, y=y, z=z)

    # rotation matrix
    R = R @ create_rotation_matrix(x=roll, y=pitch, z=yaw)
    return R


def pixel_wise_rot(M: torch.Tensor) -> Tuple[torch.Tensor]:
    r"""Rotation coordinates to phi/theta of the equirectangular image

    params:
        M: torch.Tensor

    return:
        phis: torch.Tensor
        thetas: torch.Tensor
    """
    if len(M.shape) == 3:
        M = M.unsqueeze(0)

    norms = torch.norm(M, dim=-1)
    thetas = torch.atan2(M[:, :, :, 0], M[:, :, :, 2])
    phis = torch.asin(M[:, :, :, 1] / norms)

    if thetas.shape[0] == phis.shape[0] == 1:
        thetas = thetas.squeeze(0)
        phis = phis.squeeze(0)
    return phis, thetas



#!/usr/bin/env python3

import copy
import time
import os.path as osp

import numpy as np

import torch

from PIL import Image

from torchvision import transforms

from equilib.equi2pers import TorchEqui2Pers

SAMPLING_METHOD = "torch"
SAMPLING_MODE = "bilinear"

WIDTH = 640
HEIGHT = 480
FOV = 90

DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")


def run(equi, rot) -> torch.Tensor:
    h_equi, w_equi = equi.shape[-2:]
    print("equirectangular image size:")
    print(h_equi, w_equi)

    tic = time.perf_counter()
    equi2pers = TorchEqui2Pers(w_pers=WIDTH, h_pers=HEIGHT, fov_x=FOV)
    toc = time.perf_counter()
    print("Init Equi2Pers: {:0.4f} seconds".format(toc - tic))

    tic = time.perf_counter()
    sample = equi2pers(
        equi=equi,
        rot=rot,
        sampling_method=SAMPLING_METHOD,
        mode=SAMPLING_MODE,
        debug=True,
    )
    toc = time.perf_counter()
    print("Sample: {:0.4f} seconds".format(toc - tic))

    return sample


def test_torch_single():
    data_path = osp.join(".", "tests", "data")
    result_path = osp.join(".", "tests", "results")
    equi_path = osp.join(data_path, "test.jpg")
    device = DEVICE

    # Transforms
    to_tensor = transforms.Compose(
        [
            transforms.ToTensor(),
        ]
    )

    to_PIL = transforms.Compose(
        [
            transforms.ToPILImage(),
        ]
    )

    tic = time.perf_counter()
    equi_img = Image.open(equi_path)
    # NOTE: Sometimes images are RGBA
    equi_img = equi_img.convert("RGB")
    equi = to_tensor(equi_img)
    equi = equi.to(device)
    toc = time.perf_counter()
    print("Process equirectangular image: {:0.4f} seconds".format(toc - tic))

    rot = {
        "roll": 0.0,
        "pitch": 0.0,
        "yaw": 0.0,
    }

    sample = run(equi, rot)

    tic = time.perf_counter()
    pers = sample.to("cpu")
    pers_img = to_PIL(pers)
    toc = time.perf_counter()
    print("post process: {:0.4f} seconds".format(toc - tic))

    pers_path = osp.join(result_path, "output_torch_single.jpg")
    pers_img.save(pers_path)


def test_torch_batch():
    data_path = osp.join(".", "tests", "data")
    result_path = osp.join(".", "tests", "results")
    equi_path = osp.join(data_path, "test.jpg")
    device = DEVICE
    batch_size = 16

    # Transforms
    to_tensor = transforms.Compose(
        [
            transforms.ToTensor(),
        ]
    )

    to_PIL = transforms.Compose(
        [
            transforms.ToPILImage(),
        ]
    )

    tic = time.perf_counter()
    equi_img = Image.open(equi_path)
    # NOTE: Sometimes images are RGBA
    equi_img = equi_img.convert("RGB")
    batched_equi = []
    for i in range(batch_size):
        equi = to_tensor(equi_img)
        batched_equi.append(copy.deepcopy(equi))
    batched_equi = torch.stack(batched_equi, dim=0)
    batched_equi = batched_equi.to(device)
    toc = time.perf_counter()
    print("Process equirectangular image: {:0.4f} seconds".format(toc - tic))

    batched_rot = []
    inc = np.pi / 8
    for i in range(batch_size):
        rot = {
            "roll": 0,
            "pitch": i * inc,
            "yaw": 0,
        }
        batched_rot.append(rot)

    batched_sample = run(batched_equi, batched_rot)
    tic = time.perf_counter()
    batched_pers = []
    for i in range(batch_size):
        sample = copy.deepcopy(batched_sample[i])
        sample = sample.to("cpu")
        pers_img = to_PIL(sample)
        batched_pers.append(pers_img)
    toc = time.perf_counter()
    print("post process: {:0.4f} seconds".format(toc - tic))

    for i, pers in enumerate(batched_pers):
        pers_path = osp.join(result_path, "output_torch_batch_{}.jpg".format(i))
        pers.save(pers_path)



#!/usr/bin/env python3

from .base import BaseEqui2Pers
from .equi2pers_numpy.equi2pers import Equi2Pers as NumpyEqui2Pers
from .equi2pers_torch.equi2pers import Equi2Pers as TorchEqui2Pers

__all__ = [
    "BaseEqui2Pers",
    "NumpyEqui2Pers",
    "TorchEqui2Pers",
]



#!/usr/bin/env python3


class BaseEqui2Cube(object):
    r"""Base Equi2Cube class to build off of"""

    def __init__(self, w_face: int, **kwargs) -> None:
        r"""
        params:
            w_face: cube face width (int)
        """
        self.w_face = w_face

    def __call__(self, **kwargs):
        return self.run(**kwargs)

    def run(self, **kwargs):
        raise NotImplementedError



#!/usr/bin/env python3

import math
from typing import Dict, List, Tuple, Union

import torch

from equilib.grid_sample import torch_func

from ..base import BaseCube2Equi
from .utils import get_device


def cube_list2h(cube_list: List[torch.Tensor]) -> torch.Tensor:
    assert len(cube_list) == 6
    assert sum(face.shape == cube_list[0].shape for face in cube_list) == 6
    return torch.cat(cube_list, dim=-1)


def cube_dict2h(
    cube_dict: Dict[str, torch.Tensor], face_k=["F", "R", "B", "L", "U", "D"]
) -> torch.Tensor:
    assert len(face_k) == 6
    return cube_list2h([cube_dict[k] for k in face_k])


def cube_dice2h(cube_dice: torch.Tensor) -> torch.Tensor:
    r"""dice to horizion

    params:
    cube_dice: (C, H, W)
    """
    # Order: F R B L U D
    sxy = [(1, 1), (2, 1), (3, 1), (0, 1), (1, 0), (1, 2)]
    w = cube_dice.shape[-2] // 3
    assert cube_dice.shape[-2] == w * 3 and cube_dice.shape[-1] == w * 4
    cube_h = torch.zeros((cube_dice.shape[-3], w, w * 6), dtype=cube_dice.dtype)
    for i, (sx, sy) in enumerate(sxy):
        face = cube_dice[:, sy * w : (sy + 1) * w, sx * w : (sx + 1) * w]
        cube_h[:, :, i * w : (i + 1) * w] = face
    return cube_h


class Cube2Equi(BaseCube2Equi):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def _to_horizon(
        self,
        cubemap: torch.Tensor,
        cube_format: str,
    ) -> torch.Tensor:
        if cube_format == "horizon":
            pass
        elif cube_format == "list":
            cubemap = cube_list2h(cubemap)
        elif cube_format == "dict":
            cubemap = cube_dict2h(cubemap)
        elif cube_format == "dice":
            cubemap = cube_dice2h(cubemap)
        else:
            raise NotImplementedError("unknown cube_format")

        assert len(cubemap.shape) == 3
        assert cubemap.shape[-2] * 6 == cubemap.shape[-1]

        return cubemap

    def _equirect_facetype(self, h: int, w: int) -> torch.Tensor:
        r"""0F 1R 2B 3L 4U 5D"""
        tp = torch.roll(
            torch.arange(4)  # 1
            .repeat_interleave(w // 4)  # 2 same as np.repeat
            .unsqueeze(0)
            .transpose(0, 1)  # 3
            .repeat(1, h)  # 4
            .view(-1, h)  # 5
            .transpose(0, 1),  # 6
            shifts=3 * w // 8,
            dims=1,
        )

        # Prepare ceil mask
        mask = torch.zeros((h, w // 4), dtype=torch.bool)
        idx = torch.linspace(-math.pi, math.pi, w // 4) / 4
        idx = h // 2 - torch.round(
            torch.atan(torch.cos(idx)) * h / math.pi
        ).type(torch.int)
        for i, j in enumerate(idx):
            mask[:j, i] = 1
        mask = torch.roll(torch.cat([mask] * 4, 1), 3 * w // 8, 1)

        tp[mask] = 4
        tp[torch.flip(mask, dims=(0,))] = 5

        return tp.type(torch.int32)

    def create_equi_grid(self, h_out: int, w_out: int) -> Tuple[torch.Tensor]:
        _dtype = torch.float32
        theta = torch.linspace(-math.pi, math.pi, steps=w_out, dtype=_dtype)
        phi = torch.linspace(math.pi, -math.pi, steps=h_out, dtype=_dtype) / 2
        phi, theta = torch.meshgrid([phi, theta])
        return theta, phi

    def run(
        self,
        cubemap: Union[
            torch.Tensor, Dict[str, torch.Tensor], List[torch.Tensor]
        ],
        cube_format: str = "dice",
        sampling_method: str = "torch",
        mode: str = "bilinear",
    ) -> torch.Tensor:
        r""""""

        # Convert all cubemap format to `horizon` and batched
        if cube_format in ["dice", "horizon"]:
            if isinstance(cubemap, torch.Tensor):
                assert (
                    len(cubemap.shape) >= 3
                ), "input shape {} is not valid".format(cubemap.shape)
                if len(cubemap.shape) == 4:
                    cubemap = [
                        self._to_horizon(c, cube_format) for c in cubemap
                    ]
                else:
                    cubemap = [self._to_horizon(cubemap, cube_format)]
            elif isinstance(cubemap, list):
                cubemap = [self._to_horizon(c, cube_format) for c in cubemap]
            else:
                raise ValueError
        elif cube_format == "dict":
            if isinstance(cubemap, dict):
                cubemap = [self._to_horizon(cubemap, cube_format)]
            elif isinstance(cubemap, list):
                cubemap = [self._to_horizon(c, cube_format) for c in cubemap]
            else:
                raise ValueError
        elif cube_format == "list":
            if isinstance(cubemap[0], torch.Tensor):
                cubemap = [self._to_horizon(cubemap, cube_format)]
            elif isinstance(cubemap[0], list):
                cubemap = [self._to_horizon(c, cube_format) for c in cubemap]
            else:
                raise ValueError
        else:
            raise ValueError

        batch = len(cubemap)
        cubemap = torch.stack(cubemap, 0)

        # get device
        device = get_device(cubemap)
        w_face = cubemap.shape[-2]

        theta, phi = self.create_equi_grid(self.h_out, self.w_out)

        # Get face id to each pixel: 0F 1R 2B 3L 4U 5D
        tp = self._equirect_facetype(self.h_out, self.w_out)
        coor_x = torch.zeros((self.h_out, self.w_out))
        coor_y = torch.zeros((self.h_out, self.w_out))

        # FIXME: there's a bug where left section (3L) has artifacts
        # on top and bottom
        # It might have to do with 4U or 5D
        for i in range(6):
            mask = tp == i

            if i < 4:
                coor_x[mask] = 0.5 * torch.tan(theta[mask] - math.pi * i / 2)
                coor_y[mask] = (
                    -0.5
                    * torch.tan(phi[mask])
                    / torch.cos(theta[mask] - math.pi * i / 2)
                )
            elif i == 4:
                c = 0.5 * torch.tan(math.pi / 2 - phi[mask])
                coor_x[mask] = c * torch.sin(theta[mask])
                coor_y[mask] = c * torch.cos(theta[mask])
            elif i == 5:
                c = 0.5 * torch.tan(math.pi / 2 - torch.abs(phi[mask]))
                coor_x[mask] = c * torch.sin(theta[mask])
                coor_y[mask] = -c * torch.cos(theta[mask])

        # Final renormalize
        coor_x = torch.clamp(
            torch.clamp(coor_x + 0.5, 0, 1) * w_face, 0, w_face - 1
        )
        coor_y = torch.clamp(
            torch.clamp(coor_y + 0.5, 0, 1) * w_face, 0, w_face - 1
        )

        # change x axis of the x coordinate map
        for i in range(6):
            mask = tp == i
            coor_x[mask] = coor_x[mask] + w_face * i

        # repeat batch
        coor_x = coor_x.repeat(batch, 1, 1)
        coor_y = coor_y.repeat(batch, 1, 1)

        grid = torch.stack((coor_y, coor_x), axis=-3).to(device)
        grid_sample = getattr(torch_func, sampling_method)
        equis = grid_sample(cubemap, grid, mode=mode)

        if batch == 1:
            equis = equis.squeeze(0)

        return equis



#!/usr/bin/env python3

import time
import os.path as osp

from PIL import Image

import torch

from torchvision import transforms

from equilib.cube2equi import TorchCube2Equi

OUT_W = 480
OUT_H = 240

DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")


def run(cube, cube_format) -> torch.Tensor:
    print("Input is {}".format(cube_format))
    if isinstance(cube, list):
        c = cube[0]  # get the first of the batch
        if isinstance(c, list):
            assert cube_format == "list"
            print("one: {}".format(c[0].shape))
        elif isinstance(c, dict):
            assert cube_format == "dict"
            print("one: {}".format(c["F"].shape))
        elif isinstance(c, torch.Tensor):
            assert cube_format in ["horizon", "dice", "list"]
            # can be single list
            if cube_format == "list":
                print("one: {}".format(c.shape))
            else:
                print("one: {}".format(c[0].shape))
    elif isinstance(cube, dict):
        assert cube_format == "dict"
        print("one: {}".format(cube["F"].shape))
    else:
        assert cube_format in ["horizon", "dice"]
        print("one: {}".format(cube.shape))

    tic = time.perf_counter()
    cube2equi = TorchCube2Equi(
        w_out=OUT_W,
        h_out=OUT_H,
    )
    toc = time.perf_counter()
    print("Init Cube2Equi: {:0.4f} seconds".format(toc - tic))

    tic = time.perf_counter()
    samples = cube2equi(
        cubemap=cube,
        cube_format=cube_format,
        sampling_method="torch",
        mode="bilinear",
    )
    toc = time.perf_counter()
    print("Sample: {:0.4f} seconds".format(toc - tic))

    return samples


def test_torch_single():
    data_path = osp.join(".", "tests", "data")
    result_path = osp.join(".", "tests", "results")

    cube_format = "dict"
    device = DEVICE

    # Transforms
    to_tensor = transforms.Compose(
        [
            transforms.ToTensor(),
        ]
    )

    to_PIL = transforms.Compose(
        [
            transforms.ToPILImage(),
        ]
    )

    tic = time.perf_counter()
    if cube_format in ["horizon", "dice"]:
        img_path = osp.join(data_path, "test_{}.jpg".format(cube_format))
        cube = Image.open(img_path)
        cube = cube.convert("RGB")
        cube = to_tensor(cube).to(device)
    elif cube_format in ["dict", "list"]:
        img_paths = osp.join(data_path, "test_dict_{k}.jpg")
        cube = {}
        for k in ["F", "R", "B", "L", "U", "D"]:
            face = Image.open(img_paths.format(cube_format=cube_format, k=k))
            face = face.convert("RGB")
            face = to_tensor(face).to(device)
            cube[k] = face
        if cube_format == "list":
            cube = list(cube.values())
    else:
        raise ValueError
    toc = time.perf_counter()
    print("Process Cube Image: {:0.4f} seconds".format(toc - tic))

    equi = run(cube, cube_format=cube_format)

    tic = time.perf_counter()
    equi_img = to_PIL(equi.to("cpu"))
    out_path = osp.join(
        result_path, "cube2equi_torch_single_{}.jpg".format(cube_format)
    )
    equi_img.save(out_path)
    toc = time.perf_counter()


def test_torch_batch():
    data_path = osp.join(".", "tests", "data")
    result_path = osp.join(".", "tests", "results")

    batch_size = 4
    cube_format = "dict"
    device = DEVICE

    # Transforms
    to_tensor = transforms.Compose(
        [
            transforms.ToTensor(),
        ]
    )

    to_PIL = transforms.Compose(
        [
            transforms.ToPILImage(),
        ]
    )

    tic = time.perf_counter()
    batched_cube = []
    for _ in range(batch_size):
        if cube_format in ["horizon", "dice"]:
            img_path = osp.join(data_path, "test_{}.jpg".format(cube_format))
            cube = Image.open(img_path)
            cube = cube.convert("RGB")
            cube = to_tensor(cube).to(device)
        elif cube_format in ["dict", "list"]:
            img_paths = osp.join(data_path, "test_dict_{k}.jpg")
            cube = {}
            for k in ["F", "R", "B", "L", "U", "D"]:
                face = Image.open(
                    img_paths.format(cube_format=cube_format, k=k)
                )
                face = face.convert("RGB")
                face = to_tensor(face).to(device)
                cube[k] = face
            if cube_format == "list":
                cube = list(cube.values())
        else:
            raise ValueError

        batched_cube.append(cube)
    toc = time.perf_counter()
    print("Process Cube Image: {:0.4f} seconds".format(toc - tic))

    batched_equi = run(batched_cube, cube_format=cube_format)

    tic = time.perf_counter()
    for i, equi in enumerate(batched_equi):
        equi_img = to_PIL(equi.to("cpu"))
        out_path = osp.join(
            result_path,
            "cube2equi_torch_batched_{}_{}.jpg".format(cube_format, i),
        )
        equi_img.save(out_path)
    toc = time.perf_counter()
    print("post process: {:0.4f} seconds".format(toc - tic))



#!/usr/bin/env python3

from typing import List

import torch

from .utils import get_device
from .interp import linear_interp


def interp2d(
    Q: List[torch.Tensor],
    dy: torch.Tensor,
    dx: torch.Tensor,
    mode: str = "bilinear",
) -> torch.Tensor:
    r"""Naive Interpolation
    (y,x): target pixel
    mode: interpolation mode
    """
    q00, q10, q01, q11 = Q
    if mode == "bilinear":
        f0 = linear_interp(q00, q01, dx, 1.0)
        f1 = linear_interp(q10, q11, dx, 1.0)
        return linear_interp(f0, f1, dy, 1.0)
    else:
        raise NotImplementedError


def grid_sample(
    img: torch.Tensor,
    grid: torch.Tensor,
    mode: str = "bilinear",
) -> torch.Tensor:
    r"""Torch Grid Sample (Custom)

    params:
        img: Tensor[B, C, H, W]  or Tensor[C, H, W]
        grid: Tensor[B, 2, H, W] or Tensor[2, H, W]
        mode: `bilinear` or `nearest`

    returns:
        out: Tensor[B, C, H, W] or Tensor[C, H, W]
            where H, W are grid size
    """
    assert len(img.shape) == len(grid.shape), "ERR: img and grid does not match"
    assert len(img.shape) > 2, "ERR: dim needs to be 3 or 4"
    if len(img.shape) == len(grid.shape) == 3:
        img = img.unsqueeze(0)
        grid = grid.unsqueeze(0)
    batches, channels, h_in, w_in = img.shape
    _, _, h_out, w_out = grid.shape

    assert get_device(img) == get_device(
        grid
    ), "ERR: img and grid does not match device " "{} vs {}".format(
        get_device(img), get_device(grid)
    )
    device = get_device(img)

    _dtype = img.dtype
    _max = torch.tensor(1.0, device=device)
    _min = torch.tensor(0.0, device=device)

    # Initialize output image
    out = torch.zeros(
        (batches, channels, h_out, w_out), dtype=_dtype, device=device
    )

    if mode == "bilinear":
        min_grid = torch.floor(grid).type(torch.int64)
        max_grid = min_grid + 1
        d_grid = grid - min_grid

        max_grid[:, 0, :, :] = torch.where(
            max_grid[:, 0, :, :] >= h_in,
            max_grid[:, 0, :, :] - h_in,
            max_grid[:, 0, :, :],
        )
        max_grid[:, 1, :, :] = torch.where(
            max_grid[:, 1, :, :] >= w_in,
            max_grid[:, 1, :, :] - w_in,
            max_grid[:, 1, :, :],
        )

        y_mins = min_grid[:, 0, :, :]
        x_mins = min_grid[:, 1, :, :]
        y_mins = y_mins.view(batches, -1)
        x_mins = x_mins.view(batches, -1)

        y_maxs = max_grid[:, 0, :, :]
        x_maxs = max_grid[:, 1, :, :]
        y_maxs = y_maxs.view(batches, -1)
        x_maxs = x_maxs.view(batches, -1)

        y_d = d_grid[:, 0, :, :]
        x_d = d_grid[:, 1, :, :]
        y_d = y_d.view(batches, -1)
        x_d = x_d.view(batches, -1)

        # Q00 = torch.zeros((batches, channels, y_mins.shape[-1]))
        # Q10 = torch.zeros((batches, channels, y_mins.shape[-1]))
        # Q01 = torch.zeros((batches, channels, y_mins.shape[-1]))
        # Q11 = torch.zeros((batches, channels, y_mins.shape[-1]))
        # FIXME: looping batch... slow...
        for b in range(batches):
            Q00 = img[b][:, y_mins[b], x_mins[b]]
            Q10 = img[b][:, y_maxs[b], x_mins[b]]
            Q01 = img[b][:, y_mins[b], x_maxs[b]]
            Q11 = img[b][:, y_maxs[b], x_maxs[b]]

            out[b] = interp2d(
                [Q00, Q10, Q01, Q11],
                y_d[b],
                x_d[b],
                mode=mode,
            ).view(channels, h_out, w_out)

        # NOTE: it is faster, but uses too much memory
        # def _flatten(t):
        #     return t.reshape((t.shape[0] * t.shape[1]))

        # # trying to reduce for loops
        # y_d = _flatten(y_d)
        # x_d = _flatten(x_d)
        # y_mins = _flatten(y_mins)
        # x_mins = _flatten(x_mins)
        # y_maxs = _flatten(y_maxs)
        # x_maxs = _flatten(x_maxs)

        # img = img.permute(1, 0, 2, 3)
        # img = img.view(channels, -1, h_in, w_in).squeeze(1)

        # out = interp2d(
        #     [
        #         img[:, y_mins, x_mins],
        #         img[:, y_maxs, x_mins],
        #         img[:, y_mins, x_maxs],
        #         img[:, y_maxs, x_maxs],
        #     ],
        #     y_d,
        #     x_d,
        #     mode=mode,
        # ).reshape(channels, batches, h_out, w_out).permute(1, 0, 2, 3)

    elif mode == "nearest":
        round_grid = torch.round(grid).type(torch.int64)
        y = round_grid[:, 0, :, :]
        x = round_grid[:, 1, :, :]
        y = y.view(batches, -1)
        x = x.view(batches, -1)

        for b in range(batches):
            out[b] = img[b][: y[b], x[b]]

    else:
        raise ValueError("{} is not available".format(mode))

    out = torch.where(out >= _max, _max, out)
    out = torch.where(out < _min, _min, out)
    out = out.reshape(batches, channels, h_out, w_out)
    out = out.type(_dtype)
    if out.shape[0] == 1:
        out = out.squeeze(0)
    return out



#!/usr/bin/env python3

import math
from typing import Dict, List, Tuple, Union

import torch

from equilib.grid_sample import torch_func

from ..base import BaseEqui2Cube
from .utils import (
    create_rotation_matrix,
    cube_h2dice,
    cube_h2dict,
    cube_h2list,
    get_device,
    sizeof,
)


class Equi2Cube(BaseEqui2Cube):
    r"""Equi2Cube PyTorch"""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def rotation_matrix(
        self,
        roll: float,
        pitch: float,
        yaw: float,
    ) -> torch.Tensor:
        r"""Create Rotation Matrix

        params:
            roll: x-axis rotation float
            pitch: y-axis rotation float
            yaw: z-axis rotation float

        return:
            rotation matrix: torch.Tensor
        """
        R = create_rotation_matrix(x=roll, y=pitch, z=yaw)
        return R

    def create_xyz(self, w_face: int) -> torch.Tensor:
        r"""xyz coordinates of the faces of the cube"""
        _dtype = torch.float32
        out = torch.zeros((w_face, w_face * 6, 3), dtype=_dtype)
        rng = torch.linspace(-0.5, 0.5, w_face, dtype=_dtype)

        # NOTE: https://github.com/pytorch/pytorch/issues/15301
        # Torch meshgrid behaves differently than numpy

        # Front face (x = 0.5)
        out[:, 0 * w_face : 1 * w_face, [2, 1]] = torch.stack(
            torch.meshgrid([-rng, rng]), -1
        )
        out[:, 0 * w_face : 1 * w_face, 0] = 0.5

        # Right face (y = -0.5)
        out[:, 1 * w_face : 2 * w_face, [2, 0]] = torch.stack(
            torch.meshgrid([-rng, -rng]), -1
        )
        out[:, 1 * w_face : 2 * w_face, 1] = 0.5

        # Back face (x = -0.5)
        out[:, 2 * w_face : 3 * w_face, [2, 1]] = torch.stack(
            torch.meshgrid([-rng, -rng]), -1
        )
        out[:, 2 * w_face : 3 * w_face, 0] = -0.5

        # Left face (y = 0.5)
        out[:, 3 * w_face : 4 * w_face, [2, 0]] = torch.stack(
            torch.meshgrid([-rng, rng]), -1
        )
        out[:, 3 * w_face : 4 * w_face, 1] = -0.5

        # Up face (z = 0.5)
        out[:, 4 * w_face : 5 * w_face, [0, 1]] = torch.stack(
            torch.meshgrid([rng, rng]), -1
        )
        out[:, 4 * w_face : 5 * w_face, 2] = 0.5

        # Down face (z = -0.5)
        out[:, 5 * w_face : 6 * w_face, [0, 1]] = torch.stack(
            torch.meshgrid([-rng, rng]), -1
        )
        out[:, 5 * w_face : 6 * w_face, 2] = -0.5

        return out

    def xyz2rot(self, xyz) -> Tuple[torch.Tensor]:
        r"""Return rotation (theta, phi) from xyz"""
        norm = torch.norm(xyz, dim=-1)
        phi = torch.asin(xyz[:, :, :, 2] / norm)
        theta = torch.atan2(xyz[:, :, :, 1], xyz[:, :, :, 0])
        return theta, phi

    def run(
        self,
        equi: Union[torch.Tensor, List[torch.Tensor]],
        rot: Union[Dict[str, float], List[Dict[str, float]]],
        cube_format: str,
        sampling_method: str = "torch",
        mode: str = "bilinear",
        debug: bool = False,
    ) -> Union[
        torch.Tensor,
        List[torch.Tensor],
        List[Dict[str, torch.Tensor]],
        Dict[str, torch.Tensor],
    ]:
        r"""Call Equi2Cube

        params:
            equi: Union[torch.Tensor, List[torch.Tensor]]
            rot: Union[Dict[str, float], List[Dict[str, float]]]
            cube_format: str ('list', 'dict', 'dice')
            sampling_method: str (default = 'faster')
            mode: str (default = 'bilinear')
        """

        assert type(equi) == torch.Tensor, (
            "ERR: input equi expected to be `torch.Tensor` "
            "but got {}".format(type(equi))
        )
        _original_shape_len = len(equi.shape)
        assert _original_shape_len >= 3, "ERR: got {} for input equi".format(
            _original_shape_len
        )
        if _original_shape_len == 3:
            equi = equi.unsqueeze(dim=0)
            rot = [rot]

        h_equi, w_equi = equi.shape[-2:]
        if debug:
            print("equi: ", sizeof(equi) / 10e6, "mb")

        # get device
        device = get_device(equi)

        # define variables
        xyz = []
        for r in rot:
            # for each rotations calculate M
            _xyz = self.create_xyz(self.w_face)
            xyz_ = self.rotation_matrix(**r) @ _xyz.unsqueeze(3)
            xyz_ = xyz_.squeeze(3)
            xyz.append(xyz_)
        xyz = torch.stack(xyz, dim=0).to(device)

        theta, phi = self.xyz2rot(xyz)

        # center the image and convert to pixel location
        ui = (theta - math.pi) * w_equi / (2 * math.pi)
        uj = (math.pi / 2 - phi) * h_equi / math.pi

        # out-of-bounds calculations
        ui = torch.where(ui < 0, ui + w_equi, ui)
        ui = torch.where(ui >= w_equi, ui - w_equi, ui)
        uj = torch.where(uj < 0, uj + h_equi, uj)
        uj = torch.where(uj >= h_equi, uj - h_equi, uj)
        grid = torch.stack((uj, ui), dim=-3)

        grid_sample = getattr(
            torch_func,
            sampling_method,
        )
        cubemap = grid_sample(equi, grid, mode=mode)

        if cube_format == "horizon":
            pass
        elif cube_format == "list":
            cubemap = cube_h2list(cubemap)
        elif cube_format == "dict":
            cubemap = cube_h2dict(cubemap)
        elif cube_format == "dice":
            cubemap = cube_h2dice(cubemap)
        else:
            raise NotImplementedError()

        return cubemap



#!/usr/bin/env python3

from typing import Dict, List, Tuple, Union

import numpy as np

from equilib.grid_sample import numpy_func

from ..base import BaseEqui2Pers
from .utils import create_rotation_matrix

__all__ = ["Equi2Pers"]


class Equi2Pers(BaseEqui2Pers):
    def __init__(self, **kwargs) -> None:
        r"""Equi2Pers Numpy"""
        super().__init__(**kwargs)

        # initialize intrinsic matrix
        _ = self.intrinsic_matrix
        # initialize global to camera rotation matrix
        _ = self.global2camera_rotation_matrix

    @property
    def intrinsic_matrix(self) -> np.ndarray:
        r"""Create Intrinsic Matrix

        return:
            K: 3x3 matrix numpy.ndarray

        NOTE:
            ref: http://ksimek.github.io/2013/08/13/intrinsic/
        """
        if not hasattr(self, "_K"):
            # perspective projection (focal length)
            f = self.w_pers / (2.0 * np.tan(np.radians(self.fov_x) / 2.0))
            # transform between camera frame and pixel coordinates
            self._K = np.array(
                [
                    [f, self.skew, self.w_pers / 2],
                    [0.0, f, self.h_pers / 2],
                    [0.0, 0.0, 1.0],
                ]
            )
        return self._K

    @property
    def perspective_coordinate(self) -> np.ndarray:
        r"""Create mesh coordinate grid with perspective height and width

        return:
            coordinate: numpy.ndarray
        """
        _xs = np.linspace(0, self.w_pers - 1, self.w_pers)
        _ys = np.linspace(0, self.h_pers - 1, self.h_pers)
        xs, ys = np.meshgrid(_xs, _ys)
        zs = np.ones_like(xs)
        coord = np.stack((xs, ys, zs), axis=2)
        return coord

    @property
    def global2camera_rotation_matrix(self) -> np.ndarray:
        r"""Default rotation that changes global to camera coordinates"""
        if not hasattr(self, "_g2c_rot"):
            R_XY = np.array(
                [  # X <-> Y
                    [0.0, 1.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0],
                ]
            )
            R_YZ = np.array(
                [  # Y <-> Z
                    [1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0],
                    [0.0, 1.0, 0.0],
                ]
            )
            self._g2c_rot = R_XY @ R_YZ

        return self._g2c_rot

    def rotation_matrix(
        self,
        roll: float,
        pitch: float,
        yaw: float,
    ) -> np.ndarray:
        r"""Create Rotation Matrix

        params:
            roll: x-axis rotation float
            pitch: y-axis rotation float
            yaw: z-axis rotation float

        return:
            rotation matrix: numpy.ndarray

        Camera coordinates -> z-axis points forward, y-axis points upward
        Global coordinates -> x-axis points forward, z-axis points upward
        """
        R = create_rotation_matrix(x=roll, y=pitch, z=yaw)
        return R

    @staticmethod
    def _get_img_size(img: np.ndarray) -> Tuple[int]:
        r"""Return height and width"""
        return img.shape[-2:]

    def _run_single(
        self,
        equi: np.ndarray,
        rot: Dict[str, float],
        sampling_method: str,
        mode: str,
    ) -> np.ndarray:
        # define variables
        h_equi, w_equi = self._get_img_size(equi)
        m = self.perspective_coordinate
        K = self.intrinsic_matrix
        R = self.rotation_matrix(**rot)

        # conversion:
        K_inv = np.linalg.inv(K)
        m = m[:, :, :, np.newaxis]
        M = R @ self._g2c_rot @ K_inv @ m
        M = M.squeeze(3)

        # calculate rotations per perspective coordinates
        # phi = np.arcsin(M[:, :, 1] / np.linalg.norm(M, axis=-1))
        # theta = np.arctan2(M[:, :, 0], M[:, :, 2])
        phi = np.arcsin(M[:, :, 2] / np.linalg.norm(M, axis=-1))
        theta = np.arctan2(M[:, :, 1], M[:, :, 0])

        # center the image and convert to pixel location
        ui = (theta - np.pi) * w_equi / (2 * np.pi)
        uj = (phi - np.pi / 2) * h_equi / np.pi
        # out-of-bounds calculations
        ui = np.where(ui < 0, ui + w_equi, ui)
        ui = np.where(ui >= w_equi, ui - w_equi, ui)
        uj = np.where(uj < 0, uj + h_equi, uj)
        uj = np.where(uj >= h_equi, uj - h_equi, uj)
        grid = np.stack((uj, ui), axis=0)

        # grid sample
        grid_sample = getattr(numpy_func, sampling_method, "faster")
        sampled = grid_sample(equi, grid, mode=mode)
        return sampled

    def run(
        self,
        equi: Union[np.ndarray, List[np.ndarray]],
        rot: Union[Dict[str, float], List[Dict[str, float]]],
        sampling_method: str = "faster",
        mode: str = "bilinear",
    ) -> np.ndarray:
        r"""Run Equi2Pers

        params:
            equi: equirectangular image np.ndarray[C, H, W]
            rot: Dict[str, float]
            sampling_method: str (default="faster")
            mode: str (default="bilinear")

        returns:
            pers: perspective image np.ndarray[C, H, W]

        NOTE: input can be batched [B, C, H, W] or List[np.ndarray]
        NOTE: when using batches, the output types match
        """
        _return_type = type(equi)
        _original_shape_len = len(equi.shape)
        if _return_type == np.ndarray:
            assert (
                _original_shape_len >= 3
            ), "ERR: got {} for input equi".format(_original_shape_len)
            if _original_shape_len == 3:
                equi = equi[np.newaxis, :, :, :]
                rot = [rot]

        assert len(equi) == len(
            rot
        ), "ERR: length of equi and rot differs {} vs {}".format(
            len(equi), len(rot)
        )

        samples = []
        for p, r in zip(equi, rot):
            # iterate through batches
            # TODO: batch implementation
            sample = self._run_single(
                equi=p,
                rot=r,
                sampling_method=sampling_method,
                mode=mode,
            )
            samples.append(sample)

        if _return_type == np.ndarray:
            samples = np.stack(samples, axis=0)
            if _original_shape_len == 3:
                samples = np.squeeze(samples, axis=0)

        return samples



#!/usr/bin/env python3


class BaseCube2Equi(object):
    r"""Base Cube2Equi class to build off of"""

    def __init__(self, w_out: int, h_out: int, **kwargs) -> None:
        r"""
        params:
            w_out, h_out: equirectangular image size
        """
        assert w_out % 8 == 0 and h_out % 8 == 0
        self.w_out = w_out
        self.h_out = h_out

    def __call__(self, **kwargs):
        return self.run(**kwargs)

    def run(self, **kwargs):
        raise NotImplementedError



#!/usr/bin/env python3

from typing import Dict, List

import numpy as np


def create_rotation_matrix(
    x: float,
    y: float,
    z: float,
) -> np.ndarray:
    r"""Create Rotation Matrix

    params:
        x: x-axis rotation float
        y: y-axis rotation float
        z: z-axis rotation float

    return:
        rotation matrix: numpy.ndarray
    """
    # calculate rotation about the x-axis
    R_x = np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, np.cos(x), -np.sin(x)],
            [0.0, np.sin(x), np.cos(x)],
        ]
    )
    # calculate rotation about the y-axis
    R_y = np.array(
        [
            [np.cos(y), 0.0, np.sin(y)],
            [0.0, 1.0, 0.0],
            [-np.sin(y), 0.0, np.cos(y)],
        ]
    )
    # calculate rotation about the z-axis
    R_z = np.array(
        [
            [np.cos(z), -np.sin(z), 0.0],
            [np.sin(z), np.cos(z), 0.0],
            [0.0, 0.0, 1.0],
        ]
    )

    return R_z @ R_y @ R_x


def cube_h2list(cube_h: np.ndarray) -> List[np.ndarray]:
    assert cube_h.shape[-2] * 6 == cube_h.shape[-1]
    return np.split(cube_h, 6, axis=-1)


def cube_h2dict(cube_h: np.ndarray) -> Dict[str, np.ndarray]:
    cube_list = cube_h2list(cube_h)
    return dict(
        [
            (k, cube_list[i])
            for i, k in enumerate(["F", "R", "B", "L", "U", "D"])
        ]
    )


def cube_h2dice(cube_h: np.ndarray) -> np.ndarray:
    assert cube_h.shape[-2] * 6 == cube_h.shape[-1]
    w = cube_h.shape[-2]
    cube_dice = np.zeros((cube_h.shape[0], w * 3, w * 4), dtype=cube_h.dtype)
    cube_list = cube_h2list(cube_h)
    # Order: F R B L U D
    sxy = [(1, 1), (2, 1), (3, 1), (0, 1), (1, 0), (1, 2)]
    for i, (sx, sy) in enumerate(sxy):
        cube_dice[:, sy * w : (sy + 1) * w, sx * w : (sx + 1) * w] = cube_list[
            i
        ]
    return cube_dice



#!/usr/bin/env python3

from typing import Tuple

import numpy as np


def create_coord(
    height: int,
    width: int,
) -> np.ndarray:
    r"""Create mesh coordinate grid with height and width

    `z-axis` scale is `1`

    params:
        height: int
        width: int

    return:
        coordinate: numpy.ndarray
    """
    _xs = np.linspace(0, width - 1, width)
    _ys = np.linspace(0, height - 1, height)
    xs, ys = np.meshgrid(_xs, _ys)
    zs = np.ones_like(xs)
    coord = np.stack((xs, ys, zs), axis=2)
    return coord


def create_K(
    height: int,
    width: int,
    fov_x: float,
    skew: float = 0.0,
) -> np.ndarray:
    r"""Create Intrinsic Matrix

    params:
        height: int
        width: int
        fov_x: float
        skew: float

    return:
        K: 3x3 matrix numpy.ndarray

    NOTE:
        ref: http://ksimek.github.io/2013/08/13/intrinsic/
    """
    # perspective projection (focal length)
    f = width / (2.0 * np.tan(np.radians(fov_x) / 2.0))
    # transform between camera frame and pixel coordinates
    K = np.array([[f, skew, width / 2], [0.0, f, height / 2], [0.0, 0.0, 1.0]])
    return K


def create_rotation_matrix(
    x: float,
    y: float,
    z: float,
) -> np.ndarray:
    r"""Create Rotation Matrix

    params:
        x: x-axis rotation float
        y: y-axis rotation float
        z: z-axis rotation float

    return:
        rotation matrix: numpy.ndarray
    """
    # calculate rotation about the x-axis
    R_x = np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, np.cos(x), -np.sin(x)],
            [0.0, np.sin(x), np.cos(x)],
        ]
    )
    # calculate rotation about the y-axis
    R_y = np.array(
        [
            [np.cos(y), 0.0, np.sin(y)],
            [0.0, 1.0, 0.0],
            [-np.sin(y), 0.0, np.cos(y)],
        ]
    )
    # calculate rotation about the z-axis
    R_z = np.array(
        [
            [np.cos(z), -np.sin(z), 0.0],
            [np.sin(z), np.cos(z), 0.0],
            [0.0, 0.0, 1.0],
        ]
    )

    return R_z @ R_y @ R_x


def create_rot_mat(
    roll: float,
    pitch: float,
    yaw: float,
) -> np.ndarray:
    r"""Create Rotation Matrix

    params:
        roll: x-axis rotation float
        pitch: y-axis rotation float
        yaw: z-axis rotation float

    return:
        rotation matrix: numpy.ndarray

    Camera coordinates -> z-axis points forward, y-axis points upward
    Global coordinates -> x-axis points forward, z-axis poitns upward
    """

    # default rotation that changes global to camera coordinates
    x = np.pi
    y = np.pi
    z = np.pi
    R = create_rotation_matrix(x=x, y=y, z=z)

    # rotation matrix
    R = R @ create_rotation_matrix(x=roll, y=pitch, z=yaw)
    return R


def pixel_wise_rot(M: np.ndarray) -> Tuple[np.ndarray]:
    r"""Rotation coordinates to phi/theta of the equirectangular image

    params:
        M: np.ndarray

    return:
        phis: np.ndarray
        thetas: np.ndarray
    """
    phis = np.arcsin(M[:, :, 1] / np.linalg.norm(M, axis=-1))
    thetas = np.arctan2(M[:, :, 0], M[:, :, 2])
    return phis, thetas



#!/usr/bin/env python3

from typing import Dict, List

import numpy as np

import torch

pi = torch.Tensor([3.14159265358979323846])


def sizeof(tensor: torch.Tensor) -> float:
    r"""Get the size of a tensor"""
    assert torch.is_tensor(tensor), "ERR: is not tensor"
    return tensor.element_size() * tensor.nelement()


def get_device(a: torch.Tensor) -> torch.device:
    r"""Get device of a Tensor"""
    return torch.device(a.get_device() if a.get_device() >= 0 else "cpu")


def create_rotation_matrix(
    x: float,
    y: float,
    z: float,
) -> torch.Tensor:
    r"""Create Rotation Matrix

    params:
        x: x-axis rotation float
        y: y-axis rotation float
        z: z-axis rotation float

    return:
        rotation matrix: torch.Tensor
    """
    # calculate rotation about the x-axis
    R_x = torch.tensor(
        [
            [1.0, 0.0, 0.0],
            [0.0, np.cos(x), -np.sin(x)],
            [0.0, np.sin(x), np.cos(x)],
        ],
        dtype=torch.float,
    )
    # calculate rotation about the y-axis
    R_y = torch.tensor(
        [
            [np.cos(y), 0.0, np.sin(y)],
            [0.0, 1.0, 0.0],
            [-np.sin(y), 0.0, np.cos(y)],
        ],
        dtype=torch.float,
    )
    # calculate rotation about the z-axis
    R_z = torch.tensor(
        [
            [np.cos(z), -np.sin(z), 0.0],
            [np.sin(z), np.cos(z), 0.0],
            [0.0, 0.0, 1.0],
        ],
        dtype=torch.float,
    )

    return R_z @ R_y @ R_x


def cube_h2list(cube_h: torch.Tensor) -> List[torch.Tensor]:
    assert cube_h.shape[-2] * 6 == cube_h.shape[-1]
    return torch.split(cube_h, split_size_or_sections=cube_h.shape[-2], dim=-1)


def cube_h2dict(cube_h: torch.Tensor) -> Dict[str, torch.Tensor]:
    cube_list = cube_h2list(cube_h)
    if len(cube_h.shape) == 3:
        return dict(
            [
                (k, cube_list[i])
                for i, k in enumerate(["F", "R", "B", "L", "U", "D"])
            ]
        )
    else:
        batches = cube_h.shape[0]
        ret = []
        for b in range(batches):
            ret.append(
                dict(
                    [
                        (k, cube_list[i][b])
                        for i, k in enumerate(["F", "R", "B", "L", "U", "D"])
                    ]
                )
            )
        return ret


def cube_h2dice(cube_h: torch.Tensor) -> torch.Tensor:
    w = cube_h.shape[-2]
    # Order: F R B L U D
    sxy = [(1, 1), (2, 1), (3, 1), (0, 1), (1, 0), (1, 2)]

    if len(cube_h.shape) == 3:
        batches = 1
        cube_dice = torch.zeros(
            (cube_h.shape[-3], w * 3, w * 4), dtype=cube_h.dtype
        )
        cube_list = cube_h2list(cube_h)
        for i, (sx, sy) in enumerate(sxy):
            cube_dice[
                :, sy * w : (sy + 1) * w, sx * w : (sx + 1) * w
            ] = cube_list[i]
    else:
        batches = cube_h.shape[0]
        cube_dice = torch.zeros(
            (batches, cube_h.shape[-3], w * 3, w * 4), dtype=cube_h.dtype
        )
        cube_list = cube_h2list(cube_h)
        for b in range(batches):
            for i, (sx, sy) in enumerate(sxy):
                cube_dice[
                    b, :, sy * w : (sy + 1) * w, sx * w : (sx + 1) * w
                ] = cube_list[i][b]

    return cube_dice



#!/usr/bin/env python3

from typing import Optional


class BaseEqui2Equi(object):
    r"""Base Equi2Equi class to build off of"""

    def __init__(
        self, h_out: Optional[int] = None, w_out: Optional[int] = None, **kwargs
    ) -> None:
        r""""""
        self.h_out = h_out
        self.w_out = w_out

    def __call__(self, **kwargs):
        return self.run(**kwargs)

    def create_coordinate(self, h_out: int, w_out: int):
        raise NotImplementedError

    def rotation_matrix(self, roll: float, pitch: float, yaw: float):
        raise NotImplementedError

    @staticmethod
    def _get_img_size(img):
        raise NotImplementedError

    def run(self, **kwargs):
        raise NotImplementedError



#!/usr/bin/env python3

import numpy as np


def create_rotation_matrix(
    x: float,
    y: float,
    z: float,
) -> np.ndarray:
    r"""Create Rotation Matrix

    params:
        x: x-axis rotation float
        y: y-axis rotation float
        z: z-axis rotation float

    return:
        rotation matrix: numpy.ndarray
    """
    # calculate rotation about the x-axis
    R_x = np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, np.cos(x), -np.sin(x)],
            [0.0, np.sin(x), np.cos(x)],
        ]
    )
    # calculate rotation about the y-axis
    R_y = np.array(
        [
            [np.cos(y), 0.0, np.sin(y)],
            [0.0, 1.0, 0.0],
            [-np.sin(y), 0.0, np.cos(y)],
        ]
    )
    # calculate rotation about the z-axis
    R_z = np.array(
        [
            [np.cos(z), -np.sin(z), 0.0],
            [np.sin(z), np.cos(z), 0.0],
            [0.0, 0.0, 1.0],
        ]
    )

    return R_z @ R_y @ R_x



#!/usr/bin/env python3


class BaseEqui2Pers(object):
    r"""Base Equi2Pers class to build off of"""

    def __init__(
        self, w_pers: int, h_pers: int, fov_x: float, **kwargs
    ) -> None:
        r"""
        params:
            w_pers, h_pers: perspective size
            fov_x: perspective image fov of x-axis (width direction)
        """
        self._init_params(
            w_pers=w_pers,
            h_pers=h_pers,
            fov_x=fov_x,
        )

    def _init_params(
        self,
        w_pers: int,
        h_pers: int,
        fov_x: float,
        skew: float = 0.0,
    ) -> None:
        r"""Initialize local `self` parameters"""
        self.w_pers = w_pers
        self.h_pers = h_pers
        self.fov_x = fov_x
        self.skew = skew  # skew intrinsic parameter

    def __call__(self, **kwargs):
        return self.run(**kwargs)

    @property
    def intrinsic_matrix(self):
        raise NotImplementedError

    @property
    def perspective_coordinate(self):
        raise NotImplementedError

    @property
    def global2camera_rotation_matrix(self):
        raise NotImplementedError

    def rotation_matrix(self, roll: float, pitch: float, yaw: float):
        raise NotImplementedError

    @staticmethod
    def _get_img_size(img):
        raise NotImplementedError

    def run(self, **kwargs):
        raise NotImplementedError



#!/usr/bin/env python3

from .equi2cube_numpy.equi2cube import Equi2Cube as NumpyEqui2Cube
from .equi2cube_torch.equi2cube import Equi2Cube as TorchEqui2Cube

__all__ = [
    "NumpyEqui2Cube",
    "TorchEqui2Cube",
]



#!/usr/bin/env python3

from typing import List

import numpy as np

from .interp import linear_interp


def interp2d(
    Q: List[np.ndarray],
    dy: float,
    dx: float,
    mode: str = "bilinear",
) -> np.ndarray:
    r"""Naive Interpolation
    (y,x): target pixel
    mode: interpolation mode
    """
    q00, q10, q01, q11 = Q
    if mode == "bilinear":
        f0 = linear_interp(q00, q01, dx, 1)
        f1 = linear_interp(q10, q11, dx, 1)
        return linear_interp(f0, f1, dy, 1)
    else:
        return NotImplementedError


def grid_sample(
    img: np.ndarray,
    grid: np.ndarray,
    mode: str = "bilinear",
) -> np.ndarray:
    r"""Naive grid sample algorithm"""
    channels, h_in, w_in = img.shape
    _, h_out, w_out = grid.shape

    # Image conversion values
    if img.dtype == np.uint8:
        _min = 0
        _max = 255
        _dtype = np.uint8
    elif img.dtype == np.float64:
        _min = 0.0
        _max = 1.0
        _dtype = np.float64
    else:
        raise ValueError("{} is not supported".format(img.dtype))

    # Initialize output image
    out = np.zeros((channels, h_out, w_out), dtype=_dtype)

    if mode == "bilinear":
        min_grid = np.floor(grid).astype(np.uint64)
        # NOTE: uint8 convertion causes truncation, so use uint64
        max_grid = min_grid + 1
        d_grid = grid - min_grid

        max_grid[0, :, :] = np.where(
            max_grid[0, :, :] >= h_in,
            max_grid[0, :, :] - h_in,
            max_grid[0, :, :],
        )
        max_grid[1, :, :] = np.where(
            max_grid[1, :, :] >= w_in,
            max_grid[1, :, :] - w_in,
            max_grid[1, :, :],
        )

        for y in range(h_out):
            for x in range(w_out):
                dy, dx = d_grid[:, y, x]
                y0, x0 = min_grid[:, y, x]
                y1, x1 = max_grid[:, y, x]

                q00 = img[:, y0, x0]
                q10 = img[:, y1, x0]
                q01 = img[:, y0, x1]
                q11 = img[:, y1, x1]

                out[:, y, x] = interp2d([q00, q10, q01, q11], dy, dx, mode=mode)

    elif mode == "nearest":
        round_grid = np.rint(grid).astype(np.uint64)

        for y in range(h_out):
            for x in range(w_out):
                _y, _x = round_grid[:, y, x]
                out[:, y, x] = img[:, _y, _x]

    else:
        raise ValueError("{} is not available".format(mode))

    out = np.where(out >= _max, _max, out)
    out = np.where(out < _min, _min, out)
    return out.astype(_dtype)



#!/usr/bin/env python3

import time
from typing import Union
import os.path as osp

import numpy as np

from PIL import Image

from equilib.equi2cube import NumpyEqui2Cube

WIDTH = 256


def run(equi, rot, cube_format) -> Union[list, dict, np.ndarray]:
    h_equi, w_equi = equi.shape[-2:]
    print("equirectangular image size:")
    print(h_equi, w_equi)

    w_face = WIDTH

    tic = time.perf_counter()
    equi2cube = NumpyEqui2Cube(
        w_face=w_face,
    )
    toc = time.perf_counter()
    print("Init Equi2Cube: {:0.4f} seconds".format(toc - tic))

    tic = time.perf_counter()
    samples = equi2cube(
        equi=equi,
        rot=rot,
        cube_format=cube_format,
        sampling_method="faster",
        mode="bilinear",
    )
    toc = time.perf_counter()
    print("Sample: {:0.4f} seconds".format(toc - tic))

    return samples


def test_numpy_single():
    data_path = osp.join(".", "tests", "data")
    result_path = osp.join(".", "tests", "results")
    equi_path = osp.join(data_path, "test.jpg")

    tic = time.perf_counter()
    equi_img = Image.open(equi_path)
    # Sometimes images are RGBA
    equi_img = equi_img.convert("RGB")
    equi = np.asarray(equi_img)
    equi = np.transpose(equi, (2, 0, 1))
    toc = time.perf_counter()
    print("Process Equirectangular Image: {:0.4f} seconds".format(toc - tic))

    rot = {
        "roll": np.pi / 4,
        "pitch": 0,
        "yaw": 0,
    }

    cube_format = "dice"

    cube = run(equi, rot, cube_format=cube_format)

    tic = time.perf_counter()
    if cube_format == "dict":
        print("output: dict")
        for k, c in cube.items():
            out = np.transpose(c, (1, 2, 0))
            out_img = Image.fromarray(out)
            out_path = osp.join(
                result_path, "equi2cube_numpy_single_dict_{}.jpg".format(k)
            )
            out_img.save(out_path)
    elif cube_format == "list":
        print("output: list")
        for k, c in zip(["F", "R", "B", "L", "U", "D"], cube):
            out = np.transpose(c, (1, 2, 0))
            out_img = Image.fromarray(out)
            out_path = osp.join(
                result_path, "equi2cube_numpy_single_list_{}.jpg".format(k)
            )
            out_img.save(out_path)
    elif cube_format in ["horizon", "dice"]:
        print("output: {}".format(cube_format))
        out = np.transpose(cube, (1, 2, 0))
        out_img = Image.fromarray(out)
        out_path = osp.join(
            result_path, "equi2cube_numpy_single_{}.jpg".format(cube_format)
        )
        out_img.save(out_path)
    toc = time.perf_counter()
    print("post process: {:0.4f} seconds".format(toc - tic))


def test_numpy_batch():
    data_path = osp.join(".", "tests", "data")
    result_path = osp.join(".", "tests", "results")
    equi_path = osp.join(data_path, "test.jpg")
    batch_size = 4

    tic = time.perf_counter()
    batched_equi = []
    for _ in range(batch_size):
        equi_img = Image.open(equi_path)
        # Sometimes images are RGBA
        equi_img = equi_img.convert("RGB")
        equi = np.asarray(equi_img)
        equi = np.transpose(equi, (2, 0, 1))
        batched_equi.append(equi)
    batched_equi = np.stack(batched_equi, axis=0)
    toc = time.perf_counter()
    print("Process Equirectangular Image: {:0.4f} seconds".format(toc - tic))

    batched_rot = []
    inc = np.pi / 8
    for i in range(batch_size):
        rot = {
            "roll": 0,
            "pitch": i * inc,
            "yaw": 0,
        }
        batched_rot.append(rot)

    cube_format = "dice"

    batched_cubes = run(batched_equi, batched_rot, cube_format=cube_format)

    tic = time.perf_counter()
    if cube_format == "dict":
        print("output: dict")
        for i in range(batch_size):
            for k, c in batched_cubes[i].items():
                out = np.transpose(c, (1, 2, 0))
                out_img = Image.fromarray(out)
                out_path = osp.join(
                    result_path,
                    "equi2cube_numpy_batched_dict_{}_{}.jpg".format(i, k),
                )
                out_img.save(out_path)
    elif cube_format == "list":
        print("output: list")
        for i in range(batch_size):
            for k, c in zip(["F", "R", "B", "L", "U", "D"], batched_cubes[i]):
                out = np.transpose(c, (1, 2, 0))
                out_img = Image.fromarray(out)
                out_path = osp.join(
                    result_path,
                    "equi2cube_numpy_batched_list_{}_{}.jpg".format(i, k),
                )
                out_img.save(out_path)
    elif cube_format in ["horizon", "dice"]:
        print("output: {}".format(cube_format))
        for i in range(batch_size):
            out = np.transpose(batched_cubes[i], (1, 2, 0))
            out_img = Image.fromarray(out)
            out_path = osp.join(
                result_path,
                "equi2cube_numpy_batched_{}_{}.jpg".format(cube_format, i),
            )
            out_img.save(out_path)
    toc = time.perf_counter()
    print("post process: {:0.4f} seconds".format(toc - tic))



#!/usr/bin/env python3

from typing import Dict, List, Tuple, Union

import numpy as np

from equilib.grid_sample import numpy_func

from ..base import BaseEqui2Cube
from .utils import create_rotation_matrix, cube_h2dice, cube_h2dict, cube_h2list


class Equi2Cube(BaseEqui2Cube):
    r"""Equi2Cube Numpy"""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def rotation_matrix(
        self,
        roll: float,
        pitch: float,
        yaw: float,
    ) -> np.ndarray:
        r"""Create Rotation Matrix

        params:
            roll: x-axis rotation float
            pitch: y-axis rotation float
            yaw: z-axis rotation float

        return:
            rotation matrix: numpy.ndarray

        Global coordinates -> x-axis points forward, z-axis points upward
        """
        R = create_rotation_matrix(x=roll, y=pitch, z=yaw)
        return R

    def create_xyz(self, w_face: int) -> np.ndarray:
        r"""xyz coordinates of the faces of the cube"""
        out = np.zeros((w_face, w_face * 6, 3), np.float32)
        rng = np.linspace(-0.5, 0.5, num=w_face, dtype=np.float32)

        # Front face (x = 0.5)
        out[:, 0 * w_face : 1 * w_face, [1, 2]] = np.stack(
            np.meshgrid(rng, -rng), -1
        )
        out[:, 0 * w_face : 1 * w_face, 0] = 0.5

        # Right face (y = -0.5)
        out[:, 1 * w_face : 2 * w_face, [0, 2]] = np.stack(
            np.meshgrid(-rng, -rng), -1
        )
        out[:, 1 * w_face : 2 * w_face, 1] = 0.5

        # Back face (x = -0.5)
        out[:, 2 * w_face : 3 * w_face, [1, 2]] = np.stack(
            np.meshgrid(-rng, -rng), -1
        )
        out[:, 2 * w_face : 3 * w_face, 0] = -0.5

        # Left face (y = 0.5)
        out[:, 3 * w_face : 4 * w_face, [0, 2]] = np.stack(
            np.meshgrid(rng, -rng), -1
        )
        out[:, 3 * w_face : 4 * w_face, 1] = -0.5

        # Up face (z = 0.5)
        out[:, 4 * w_face : 5 * w_face, [1, 0]] = np.stack(
            np.meshgrid(rng, rng), -1
        )
        out[:, 4 * w_face : 5 * w_face, 2] = 0.5

        # Down face (z = -0.5)
        out[:, 5 * w_face : 6 * w_face, [1, 0]] = np.stack(
            np.meshgrid(rng, -rng), -1
        )
        out[:, 5 * w_face : 6 * w_face, 2] = -0.5

        return out

    def xyz2rot(self, xyz) -> Tuple[np.ndarray]:
        r"""Return rotation (theta, phi) from xyz"""
        norm = np.linalg.norm(xyz, axis=-1)
        phi = np.arcsin(xyz[:, :, 2] / norm)
        theta = np.arctan2(xyz[:, :, 1], xyz[:, :, 0])
        return theta, phi

    def _run_single(
        self,
        equi: np.ndarray,
        rot: Dict[str, float],
        cube_format: str,
        sampling_method: str = "faster",
        mode: str = "bilinear",
    ) -> np.ndarray:
        r"""Call a single run

        params:
            equi: np.ndarray
            rot: Dict[str, float]
            cube_format: str (default = "faster")
            mode: str (default = "bilinear")
        """

        assert len(equi.shape) == 3, "ERR: {} is not a valid array".format(
            equi.shape
        )
        assert equi.shape[0] == 3, "ERR: got {} for channel size".format(
            equi.shape[0]
        )
        h_equi, w_equi = equi.shape[-2:]

        xyz = self.create_xyz(self.w_face)
        xyz = xyz[:, :, :, np.newaxis]
        xyz_ = self.rotation_matrix(**rot) @ xyz
        xyz_ = xyz_.squeeze(-1)
        theta, phi = self.xyz2rot(xyz_)

        # center the image and convert to pixel location
        ui = (theta - np.pi) * w_equi / (2 * np.pi)
        uj = (np.pi / 2 - phi) * h_equi / np.pi

        # out-of-bounds calculations
        ui = np.where(ui < 0, ui + w_equi, ui)
        ui = np.where(ui >= w_equi, ui - w_equi, ui)
        uj = np.where(uj < 0, uj + h_equi, uj)
        uj = np.where(uj >= h_equi, uj - h_equi, uj)
        grid = np.stack((uj, ui), axis=0)

        grid_sample = getattr(numpy_func, sampling_method, "faster")
        cubemap = grid_sample(equi, grid, mode=mode)

        if cube_format == "horizon":
            pass
        elif cube_format == "list":
            cubemap = cube_h2list(cubemap)
        elif cube_format == "dict":
            cubemap = cube_h2dict(cubemap)
        elif cube_format == "dice":
            cubemap = cube_h2dice(cubemap)
        else:
            raise NotImplementedError("{} is not supported".format(cube_format))

        return cubemap

    def run(
        self,
        equi: Union[np.ndarray, List[np.ndarray]],
        rot: Union[Dict[str, float], List[Dict[str, float]]],
        cube_format: str,
        sampling_method: str = "faster",
        mode: str = "bilinear",
    ) -> Union[np.ndarray, List[np.ndarray], List[Dict[str, np.ndarray]]]:
        r"""Call Equi2Cube

        params:
            equi: Union[np.ndarray, List[np.ndarray]]
            rot: Union[Dict[str, float], List[Dict[str, float]]]
            cube_format: str ('list', 'dict', 'dice')
            sampling_method: str (default = 'faster')
            mode: str (default = 'bilinear')
        """

        _return_type = type(equi)
        _original_shape_len = len(equi.shape)
        if _return_type == np.ndarray:
            assert (
                _original_shape_len >= 3
            ), "ERR: got {} for input equi".format(_original_shape_len)
            if _original_shape_len == 3:
                equi = equi[np.newaxis, :, :, :]
                rot = [rot]

        assert len(equi) == len(
            rot
        ), "ERR: length of input and rot differs {} vs {}".format(
            len(equi), len(rot)
        )

        cubemaps = []
        for e, r in zip(equi, rot):
            # iterate through batches
            # TODO: batch implementation
            cubemap = self._run_single(
                equi=e,
                rot=r,
                cube_format=cube_format,
                sampling_method=sampling_method,
                mode=mode,
            )
            cubemaps.append(cubemap)

        if _return_type == np.ndarray:
            if cube_format in ["horizon", "dice"]:
                cubemaps = np.stack(cubemaps, axis=0)

        if _original_shape_len == 3:
            cubemaps = cubemaps[0]

        return cubemaps



#!/usr/bin/env python3

import time
from typing import Union
import os.path as osp

import numpy as np

import torch

from PIL import Image

from torchvision import transforms

from equilib.equi2cube import TorchEqui2Cube

WIDTH = 256
DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")


def run(equi, rot, cube_format) -> Union[dict, list, torch.Tensor]:
    h_equi, w_equi = equi.shape[-2:]
    print("equirectangular image size:")
    print(h_equi, w_equi)

    w_face = WIDTH

    tic = time.perf_counter()
    equi2cube = TorchEqui2Cube(
        w_face=w_face,
    )
    toc = time.perf_counter()
    print("Init Equi2Cube: {:0.4f} seconds".format(toc - tic))

    tic = time.perf_counter()
    samples = equi2cube(
        equi=equi,
        rot=rot,
        cube_format=cube_format,
        sampling_method="torch",
        mode="bilinear",
    )
    toc = time.perf_counter()
    print("Sample: {:0.4f} seconds".format(toc - tic))

    return samples


def test_torch_single():
    data_path = osp.join(".", "tests", "data")
    result_path = osp.join(".", "tests", "results")
    equi_path = osp.join(data_path, "test.jpg")
    device = DEVICE

    # Transforms
    to_tensor = transforms.Compose(
        [
            transforms.ToTensor(),
        ]
    )

    to_PIL = transforms.Compose(
        [
            transforms.ToPILImage(),
        ]
    )

    tic = time.perf_counter()
    equi_img = Image.open(equi_path)
    # Sometimes images are RGBA
    equi_img = equi_img.convert("RGB")
    equi = to_tensor(equi_img)
    equi = equi.to(device)
    toc = time.perf_counter()
    print("Process Equirectangular Image: {:0.4f} seconds".format(toc - tic))

    rot = {
        "roll": 0,
        "pitch": 0,
        "yaw": 0,
    }

    cube_format = "dice"

    cube = run(equi, rot, cube_format=cube_format)

    tic = time.perf_counter()
    if cube_format == "dict":
        print("output: dict")
        for k, c in cube.items():
            out = c.to("cpu")
            out_img = to_PIL(out)
            out_path = osp.join(
                result_path, "equi2cube_torch_single_dict_{}.jpg".format(k)
            )
            out_img.save(out_path)
    elif cube_format == "list":
        print("output: list")
        for k, c in zip(["F", "R", "B", "L", "U", "D"], cube):
            out = c.to("cpu")
            out_img = to_PIL(out)
            out_path = osp.join(
                result_path, "equi2cube_torch_single_list_{}.jpg".format(k)
            )
            out_img.save(out_path)
    elif cube_format in ["horizon", "dice"]:
        print("output: {}".format(cube_format))
        out = cube.to("cpu")
        out_img = to_PIL(out)
        out_path = osp.join(
            result_path, "equi2cube_torch_single_{}.jpg".format(cube_format)
        )
        out_img.save(out_path)
    toc = time.perf_counter()
    print("post process: {:0.4f} seconds".format(toc - tic))


def test_torch_batch():
    data_path = osp.join(".", "tests", "data")
    result_path = osp.join(".", "tests", "results")
    equi_path = osp.join(data_path, "test.jpg")
    batch_size = 4
    device = DEVICE

    # Transforms
    to_tensor = transforms.Compose(
        [
            transforms.ToTensor(),
        ]
    )

    to_PIL = transforms.Compose(
        [
            transforms.ToPILImage(),
        ]
    )

    tic = time.perf_counter()
    batched_equi = []
    for _ in range(batch_size):
        equi_img = Image.open(equi_path)
        # Sometimes images are RGBA
        equi = to_tensor(equi_img)
        batched_equi.append(equi)
    batched_equi = torch.stack(batched_equi, axis=0)
    batched_equi = batched_equi.to(device)
    toc = time.perf_counter()
    print("Process Equirectangular Image: {:0.4f} seconds".format(toc - tic))

    batched_rot = []
    inc = np.pi / 8
    for i in range(batch_size):
        rot = {
            "roll": 0,
            "pitch": i * inc,
            "yaw": 0,
        }
        batched_rot.append(rot)

    cube_format = "dict"

    batched_cubes = run(batched_equi, batched_rot, cube_format=cube_format)

    tic = time.perf_counter()
    if cube_format == "dict":
        print("output: dict")
        for i in range(batch_size):
            for k, c in batched_cubes[i].items():
                out = c.to("cpu")
                out_img = to_PIL(out)
                out_path = osp.join(
                    result_path,
                    "equi2cube_torch_batched_dict_{}_{}.jpg".format(i, k),
                )
                out_img.save(out_path)
    elif cube_format == "list":
        print("output: list")
        for i in range(batch_size):
            for k, c in zip(["F", "R", "B", "L", "U", "D"], batched_cubes[i]):
                out = c.to("cpu")
                out_img = to_PIL(out)
                out_path = osp.join(
                    result_path,
                    "equi2cube_torch_batched_list_{}_{}.jpg".format(i, k),
                )
                out_img.save(out_path)
    elif cube_format in ["horizon", "dice"]:
        print("output: {}".format(cube_format))
        for i in range(batch_size):
            out = batched_cubes[i].to("cpu")
            out_img = to_PIL(out)
            out_path = osp.join(
                result_path,
                "equi2cube_torch_batched_{}_{}.jpg".format(cube_format, i),
            )
            out_img.save(out_path)
    toc = time.perf_counter()
    print("post process: {:0.4f} seconds".format(toc - tic))



#!/usr/bin/env python3

import torch

pi = torch.Tensor([3.14159265358979323846])


def sizeof(tensor: torch.Tensor) -> float:
    r"""Get the size of a tensor"""
    assert torch.is_tensor(tensor), "ERR: is not tensor"
    return tensor.element_size() * tensor.nelement()


def get_device(a: torch.Tensor) -> torch.device:
    r"""Get device of a Tensor"""
    return torch.device(a.get_device() if a.get_device() >= 0 else "cpu")



#!/usr/bin/env python3

import distutils.spawn
import os.path as osp
import shlex
import subprocess
import sys

from setuptools import find_packages, setup


def readme():
    with open("README.md") as f:
        content = f.read()
    return content


def find_version():
    version_file = "equilib/__init__.py"
    with open(version_file, "r") as f:
        exec(compile(f.read(), version_file, "exec"))
    return locals()["__version__"]


def get_requirements(filename="requirements.txt"):
    here = osp.dirname(osp.realpath(__file__))
    with open(osp.join(here, filename), "r") as f:
        requires = [line.replace("\n", "") for line in f.readlines()]
    return requires


def get_long_description():
    with open("README.md") as f:
        long_description = f.read()

    try:
        import github2pypi

        return github2pypi.replace_url(
            slug="haruishi43/equilib", content=long_description
        )
    except Exception:
        return long_description


if sys.argv[1] == "release":
    if not distutils.spawn.find_executable("twine"):
        print("Please install twine:\n\n\tpip install twine\n", file=sys.stderr)
        sys.exit(1)

    commands = [
        "git pull origin master",
        "git tag v{:s}".format(find_version()),
        "git push origin master --tag",
        "python setup.py sdist",
        "twine upload dist/pyequilib-{:s}.tar.gz".format(find_version()),
    ]
    for cmd in commands:
        subprocess.check_call(shlex.split(cmd))
    sys.exit(0)


setup(
    name="pyequilib",
    version=find_version(),
    packages=find_packages(exclude=["github2pypi"]),
    description="equirectangular image processing with python",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author="Haruya Ishikawa",
    author_email="www.haru.ishi43@gmail.com",
    license="MIT",
    url="https://github.com/haruishi43/equilib",
    install_requires=get_requirements(),
    keywords=["Equirectangular", "Computer Vision"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    python_requires=">=3.5",
)



#!/usr/bin/env python3

from typing import Dict, List, Tuple, Union

import numpy as np

from equilib.grid_sample import numpy_func

from ..base import BaseEqui2Equi
from .utils import create_rotation_matrix

__all__ = ["Equi2Equi"]


class Equi2Equi(BaseEqui2Equi):
    def __init__(self, **kwargs) -> None:
        r"""Equi2Equi Numpy"""
        super().__init__(**kwargs)

    def create_coordinate(self, h_out: int, w_out: int) -> np.ndarray:
        r"""Create mesh coordinate grid with height and width

        return:
            coordinate: numpy.ndarray
        """
        xs = np.linspace(0, w_out - 1, w_out)
        theta = xs * 2 * np.pi / w_out - np.pi
        ys = np.linspace(0, h_out - 1, h_out)
        phi = ys * np.pi / h_out - np.pi / 2
        theta, phi = np.meshgrid(theta, phi)
        coord = np.stack((theta, phi), axis=-1)
        return coord

    def rotation_matrix(
        self,
        roll: float,
        pitch: float,
        yaw: float,
    ) -> np.ndarray:
        r"""Create Rotation Matrix

        params:
            roll: x-axis rotation float
            pitch: y-axis rotation float
            yaw: z-axis rotation float

        return:
            rotation matrix: numpy.ndarray

        Global coordinates -> x-axis points forward, z-axis points upward
        """
        R = create_rotation_matrix(x=roll, y=pitch, z=yaw)
        return R

    @staticmethod
    def _get_img_size(img: np.ndarray) -> Tuple[int]:
        r"""Return height and width"""
        return img.shape[-2:]

    def _run_single(
        self,
        src: np.ndarray,
        rot: Dict[str, float],
        sampling_method: str,
        mode: str,
    ) -> np.ndarray:
        # define variables
        h_equi, w_equi = self._get_img_size(src)
        if self.h_out is None and self.w_out is None:
            self.h_out = h_equi
            self.w_out = w_equi

        a = self.create_coordinate(self.h_out, self.w_out)  # (theta, phi)s
        norm_A = 1
        x = norm_A * np.cos(a[:, :, 1]) * np.cos(a[:, :, 0])
        y = norm_A * np.cos(a[:, :, 1]) * np.sin(a[:, :, 0])
        z = norm_A * np.sin(a[:, :, 1])
        A = np.stack((x, y, z), axis=-1)

        R = self.rotation_matrix(**rot)

        # conversion:
        # B = R @ A
        A = A[:, :, :, np.newaxis]
        B = R @ A
        B = B.squeeze(3)

        # calculate rotations per perspective coordinates
        phi = np.arcsin(B[:, :, 2] / np.linalg.norm(B, axis=-1))
        theta = np.arctan2(B[:, :, 1], B[:, :, 0])

        # center the image and convert to pixel location
        ui = (theta - np.pi) * w_equi / (2 * np.pi)
        uj = (phi - np.pi / 2) * h_equi / np.pi
        # out-of-bounds calculations
        ui = np.where(ui < 0, ui + w_equi, ui)
        ui = np.where(ui >= w_equi, ui - w_equi, ui)
        uj = np.where(uj < 0, uj + h_equi, uj)
        uj = np.where(uj >= h_equi, uj - h_equi, uj)
        grid = np.stack((uj, ui), axis=0)

        # grid sample
        grid_sample = getattr(numpy_func, sampling_method, "faster")
        sampled = grid_sample(src, grid, mode=mode)
        return sampled

    def run(
        self,
        src: Union[np.ndarray, List[np.ndarray]],
        rot: Union[Dict[str, float], List[Dict[str, float]]],
        sampling_method: str = "faster",
        mode: str = "bilinear",
    ) -> np.ndarray:
        r"""Run Equi2Pers

        params:
            src: equirectangular image np.ndarray[C, H, W]
            rot: Dict[str, float]
            sampling_method: str (default="faster")
            mode: str (default="bilinear")

        returns:
            pers: perspective image np.ndarray[C, H, W]

        NOTE: input can be batched [B, C, H, W] or List[np.ndarray]
        NOTE: when using batches, the output types match
        """
        _return_type = type(src)
        _original_shape_len = len(src.shape)
        if _return_type == np.ndarray:
            assert (
                _original_shape_len >= 3
            ), "ERR: got {} for input equi".format(_original_shape_len)
            if _original_shape_len == 3:
                src = src[np.newaxis, :, :, :]
                rot = [rot]

        assert len(src) == len(
            rot
        ), "ERR: length of src and rot differs {} vs {}".format(
            len(src), len(rot)
        )

        samples = []
        for s, r in zip(src, rot):
            # iterate through batches
            # TODO: batch implementation
            sample = self._run_single(
                src=s,
                rot=r,
                sampling_method=sampling_method,
                mode=mode,
            )
            samples.append(sample)

        if _return_type == np.ndarray:
            samples = np.stack(samples, axis=0)
            if _original_shape_len == 3:
                samples = np.squeeze(samples, axis=0)

        return samples



#!/usr/bin/env python3

import math
from typing import Dict, List, Tuple, Union

import torch

from equilib.grid_sample import torch_func

from ..base import BaseEqui2Pers
from .utils import create_rotation_matrix, deg2rad, get_device, sizeof

__all___ = ["Equi2Pers"]


class Equi2Pers(BaseEqui2Pers):
    def __init__(self, **kwargs) -> None:
        r"""Equi2Pers PyTorch"""
        super().__init__(**kwargs)

        # initialize intrinsic matrix
        _ = self.intrinsic_matrix
        # initialize global to camera rotation matrix
        _ = self.global2camera_rotation_matrix

    @property
    def intrinsic_matrix(self) -> torch.Tensor:
        r"""Create Intrinsic Matrix

        return:
            K: 3x3 matrix torch.Tensor

        NOTE:
            ref: http://ksimek.github.io/2013/08/13/intrinsic/
        """
        if not hasattr(self, "_K"):
            fov_x = torch.tensor(self.fov_x)
            f = self.w_pers / (2 * torch.tan(deg2rad(fov_x) / 2))
            self._K = torch.tensor(
                [
                    [f, self.skew, self.w_pers / 2],
                    [0.0, f, self.h_pers / 2],
                    [0.0, 0.0, 1.0],
                ]
            )
        return self._K

    @property
    def perspective_coordinate(self) -> torch.Tensor:
        r"""Create mesh coordinate grid with perspective height and width

        return:
            coordinate: torch.Tensor
        """
        _xs = torch.linspace(0, self.w_pers - 1, self.w_pers)
        _ys = torch.linspace(0, self.h_pers - 1, self.h_pers)
        # NOTE: https://github.com/pytorch/pytorch/issues/15301
        # Torch meshgrid behaves differently than numpy
        ys, xs = torch.meshgrid([_ys, _xs])
        zs = torch.ones_like(xs)
        coord = torch.stack((xs, ys, zs), dim=2)
        return coord

    @property
    def global2camera_rotation_matrix(self) -> torch.Tensor:
        r"""Default rotation that changes global to camera coordinates"""
        if not hasattr(self, "_g2c_rot"):
            R_XY = torch.tensor(
                [  # X <-> Y
                    [0.0, 1.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0],
                ]
            )
            R_YZ = torch.tensor(
                [  # Y <-> Z
                    [1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0],
                    [0.0, 1.0, 0.0],
                ]
            )
            self._g2c_rot = R_XY @ R_YZ
        return self._g2c_rot

    def rotation_matrix(
        self,
        roll: float,
        pitch: float,
        yaw: float,
    ) -> torch.Tensor:
        r"""Create Rotation Matrix

        params:
            roll: x-axis rotation float
            pitch: y-axis rotation float
            yaw: z-axis rotation float

        return:
            rotation matrix: torch.Tensor

        Camera coordinates -> z-axis points forward, y-axis points upward
        Global coordinates -> x-axis points forward, z-axis poitns upward
        """
        R = create_rotation_matrix(x=roll, y=pitch, z=yaw)
        return R

    @staticmethod
    def _get_img_size(img: torch.Tensor) -> Tuple[int]:
        r"""Return height and width"""
        # batch, channel, height, width
        return img.shape[-2:]

    def run(
        self,
        equi: torch.Tensor,
        rot: Union[Dict[str, float], List[Dict[str, float]]],
        sampling_method: str = "torch",
        mode: str = "bilinear",
        debug: bool = False,
    ) -> torch.Tensor:
        r"""Run Equi2Pers

        params:
            equi: equirectangular image torch.Tensor[(B), C, H, W]
            rot: Dict[str, float] or List[Dict[str, float]]
            sampling_method: str (default="torch")
            mode: str (default="bilinear")

        returns:
            pers: perspective image torch.Tensor[C, H, W]

        NOTE: input can be batched [B, C, H, W] or single [C, H, W]
        NOTE: when using batches, the output types match
        """
        assert type(equi) == torch.Tensor, (
            "ERR: input equi expected to be `torch.Tensor` "
            "but got {}".format(type(equi))
        )
        _original_shape_len = len(equi.shape)
        assert _original_shape_len >= 3, "ERR: got {} for input equi".format(
            _original_shape_len
        )
        if _original_shape_len == 3:
            equi = equi.unsqueeze(dim=0)
            rot = [rot]

        h_equi, w_equi = self._get_img_size(equi)
        if debug:
            print("equi: ", sizeof(equi) / 10e6, "mb")

        # get device
        device = get_device(equi)

        # define variables
        M = []
        for r in rot:
            # for each rotations calculate M
            m = self.perspective_coordinate
            K = self.intrinsic_matrix
            R = self.rotation_matrix(**r)
            _M = R @ self._g2c_rot @ K.inverse() @ m.unsqueeze(3)
            _M = _M.squeeze(3)
            M.append(_M)
        M = torch.stack(M, dim=0).to(device)

        # calculate rotations per perspective coordinates
        norms = torch.norm(M, dim=-1)
        phi = torch.asin(M[:, :, :, 2] / norms)
        theta = torch.atan2(M[:, :, :, 1], M[:, :, :, 0])

        # center the image and convert to pixel locatio
        ui = (theta - math.pi) * w_equi / (2 * math.pi)
        uj = (phi - math.pi / 2) * h_equi / math.pi
        # out-of-bounds calculations
        ui = torch.where(ui < 0, ui + w_equi, ui)
        ui = torch.where(ui >= w_equi, ui - w_equi, ui)
        uj = torch.where(uj < 0, uj + h_equi, uj)
        uj = torch.where(uj >= h_equi, uj - h_equi, uj)
        grid = torch.stack((uj, ui), axis=-3)  # 3rd to last

        # grid sample
        grid_sample = getattr(torch_func, sampling_method, "torch")
        samples = grid_sample(equi, grid, mode=mode)

        if _original_shape_len == 3:
            samples = samples.squeeze(axis=0)

        return samples



#!/usr/bin/env python3

from .cube2equi_numpy.cube2equi import Cube2Equi as NumpyCube2Equi
from .cube2equi_torch.cube2equi import Cube2Equi as TorchCube2Equi

__all__ = [
    "NumpyCube2Equi",
    "TorchCube2Equi",
]



#!/usr/bin/env python3

import torch
import torch.nn.functional as F


def grid_sample(
    img: torch.Tensor, grid: torch.Tensor, mode: str = "bilinear", **kwargs
) -> torch.Tensor:
    r"""Torch Grid Sample (default)
        Uses `torch.nn.functional.grid_sample`
        By far the best way to sample

    params:
        img: Tensor[B, C, H, W]  or Tensor[C, H, W]
        grid: Tensor[B, 2, H, W] or Tensor[2, H, W]
        device: torch.device
        mode: `bilinear` or `nearest`

    returns:
        out: Tensor[B, C, H, W] or Tensor[C, H, W]
            where H, W are grid size
    """
    if len(img.shape) == 3:
        img = img.unsqueeze(0)
        grid = grid.unsqueeze(0)

    # grid in shape: (batch, channel, h_out, w_out)
    # grid out shape: (batch, h_out, w_out, channel)
    batches, channels, h_img, w_img = img.shape
    _batches, xy, h_out, w_out = grid.shape
    assert batches == _batches, "Batch size mismatch"

    # normalize grid -1~1
    grid = grid.permute(0, 2, 3, 1)
    norm_uj = torch.clamp(
        2 * (grid[:, :, :, 0] - h_img / 2 + 0.5) / h_img, -1, 1
    )
    norm_ui = torch.clamp(
        2 * (grid[:, :, :, 1] - w_img / 2 + 0.5) / w_img, -1, 1
    )
    # grid sample takes xy, not (height, width)
    grid[:, :, :, 0] = norm_ui
    grid[:, :, :, 1] = norm_uj

    out = F.grid_sample(
        img,
        grid,
        mode=mode,
        align_corners=True,
    )
    if out.shape[0] == 1:
        out = out.squeeze(0)
    return out



#!/usr/bin/env python3

import time
import os.path as osp

import numpy as np

from PIL import Image

from equilib.equi2pers import NumpyEqui2Pers

SAMPLING_METHOD = "faster"
SAMPLING_MODE = "bilinear"

WIDTH = 640
HEIGHT = 480
FOV = 90


def run(equi, rot) -> np.ndarray:
    h_equi, w_equi = equi.shape[-2:]
    print("equirectangular image size:")
    print(h_equi, w_equi)

    tic = time.perf_counter()
    equi2pers = NumpyEqui2Pers(
        w_pers=WIDTH,
        h_pers=HEIGHT,
        fov_x=FOV,
    )
    toc = time.perf_counter()
    print("Init Equi2Pers: {:0.4f} seconds".format(toc - tic))

    tic = time.perf_counter()
    sample = equi2pers(
        equi=equi,
        rot=rot,
        sampling_method=SAMPLING_METHOD,
        mode=SAMPLING_MODE,
    )
    toc = time.perf_counter()
    print("Sample: {:0.4f} seconds".format(toc - tic))

    return sample


def test_numpy_single():
    data_path = osp.join(".", "tests", "data")
    result_path = osp.join(".", "tests", "results")
    equi_path = osp.join(data_path, "test.jpg")

    tic = time.perf_counter()
    equi_img = Image.open(equi_path)
    # Sometimes images are RGBA
    equi_img = equi_img.convert("RGB")
    equi = np.asarray(equi_img)
    equi = np.transpose(equi, (2, 0, 1))
    toc = time.perf_counter()
    print("Process Equirectangular Image: {:0.4f} seconds".format(toc - tic))

    rot = {
        "roll": 0,
        "pitch": 0,
        "yaw": 0,
    }

    sample = run(equi, rot)

    tic = time.perf_counter()
    pers = np.transpose(sample, (1, 2, 0))
    pers_img = Image.fromarray(pers)
    toc = time.perf_counter()
    print("post process: {:0.4f} seconds".format(toc - tic))

    pers_path = osp.join(result_path, "equi2pers_numpy_single.jpg")
    pers_img.save(pers_path)


def test_numpy_batch():
    data_path = osp.join(".", "tests", "data")
    result_path = osp.join(".", "tests", "results")
    equi_path = osp.join(data_path, "test.jpg")
    batch_size = 4

    tic = time.perf_counter()
    batched_equi = []
    for _ in range(batch_size):
        equi_img = Image.open(equi_path)
        # Sometimes images are RGBA
        equi_img = equi_img.convert("RGB")
        equi = np.asarray(equi_img)
        equi = np.transpose(equi, (2, 0, 1))
        batched_equi.append(equi)
    batched_equi = np.stack(batched_equi, axis=0)
    toc = time.perf_counter()
    print("Process Equirectangular Image: {:0.4f} seconds".format(toc - tic))

    batched_rot = []
    inc = np.pi / 8
    for i in range(batch_size):
        rot = {
            "roll": 0,
            "pitch": i * inc,
            "yaw": 0,
        }
        batched_rot.append(rot)

    batched_sample = run(batched_equi, batched_rot)

    tic = time.perf_counter()
    batched_pers = []
    for i in range(batch_size):
        sample = batched_sample[i]
        pers = np.transpose(sample, (1, 2, 0))
        pers_img = Image.fromarray(pers)
        batched_pers.append(pers_img)
    toc = time.perf_counter()

    for i, pers in enumerate(batched_pers):
        pers_path = osp.join(
            result_path, "equi2pers_numpy_batch_{}.jpg".format(i)
        )
        pers.save(pers_path)
